from __future__ import annotations
import functools
from io import StringIO
from math import pi, sqrt
import ase.units as units
import numpy as np
from ase.data import chemical_symbols

from gpaw import debug
from gpaw.basis_data import Basis
from gpaw.gaunt import gaunt, nabla
from gpaw.overlap import OverlapCorrections
from gpaw.setup_data import SetupData, search_for_file
from gpaw.spline import Spline
from gpaw.utilities import pack, unpack
from gpaw.xc import XC
from gpaw.new import zips
from gpaw.xc.ri.spherical_hse_kernel import RadialHSE
from gpaw.core.atom_arrays import AtomArraysLayout


class WrongMagmomForHundsRuleError(ValueError):
    """
    Custom error for catching bad magnetic moments in Hund's rule calculation
    """


def create_setup(symbol, xc='LDA', lmax=0,
                 type='paw', basis=None, setupdata=None,
                 filter=None, world=None):
    if isinstance(xc, str):
        xc = XC(xc)

    if isinstance(type, str) and ':' in type:
        from gpaw.hubbard import parse_hubbard_string
        type, hubbard_u = parse_hubbard_string(type)
    else:
        hubbard_u = None

    if setupdata is None:
        if type == 'hgh' or type == 'hgh.sc':
            lmax = 0
            from gpaw.hgh import HGHSetupData, sc_setups, setups
            if type == 'hgh.sc':
                table = sc_setups
            else:
                table = setups
            parameters = table[symbol]
            setupdata = HGHSetupData(parameters)
        elif type == 'ah':
            from gpaw.ah import AppelbaumHamann
            ah = AppelbaumHamann()
            ah.build(basis)
            return ah
        elif type == 'ae':
            from gpaw.ae import HydrogenAllElectronSetup
            assert symbol == 'H'
            ae = HydrogenAllElectronSetup()
            ae.build(basis)
            return ae
        elif type == 'ghost':
            from gpaw.lcao.bsse import GhostSetupData
            setupdata = GhostSetupData(symbol)
        elif type == 'sg15':
            from gpaw.upf import read_sg15
            upfname = f'{symbol}_ONCV_PBE-*.upf'
            try:
                upfpath, source = search_for_file(upfname, world=world)
            except RuntimeError:
                raise IOError('Could not find pseudopotential file %s '
                              'in any GPAW search path.  '
                              'Please install the SG15 setups using, '
                              'e.g., "gpaw install-data".' % upfname)
            setupdata = read_sg15(upfpath)
            if xc.get_setup_name() != 'PBE':
                raise ValueError('SG15 pseudopotentials support only the PBE '
                                 'functional.  This calculation would use '
                                 'the %s functional.' % xc.get_setup_name())
        else:
            setupdata = SetupData(symbol, xc.get_setup_name(),
                                  type, True,
                                  world=world)
    if hasattr(setupdata, 'build'):
        # It is not so nice that we have hubbard_u floating around here.
        # For example, none of the other setup types are aware
        # of hubbard u, so they silently ignore it!
        setup = LeanSetup(setupdata.build(xc, lmax, basis, filter),
                          hubbard_u=hubbard_u)
        return setup
    else:
        return setupdata


def correct_occ_numbers(f_j,
                        degeneracy_j,
                        jsorted,
                        correction: float,
                        eps=1e-12) -> None:
    """Correct f_j ndarray in-place."""

    if correction > 0:
        # Add electrons to the lowest eigenstates:
        for j in jsorted:
            c = min(correction, degeneracy_j[j] - f_j[j])
            f_j[j] += c
            correction -= c
            if correction < eps:
                break
    elif correction < 0:
        # Add electrons to the highest eigenstates:
        for j in jsorted[::-1]:
            c = min(-correction, f_j[j])
            f_j[j] -= c
            correction += c
            if correction > -eps:
                break


class LocalCorrectionVar:
    """Class holding data for local the calculation of local corr."""
    def __init__(self, s=None):
        """Initialize our data."""
        for work_key in ('nq', 'lcut', 'n_qg', 'nt_qg', 'nc_g', 'nct_g',
                         'rgd2', 'Delta_lq', 'T_Lqp'):
            if s is None or not hasattr(s, work_key):
                setattr(self, work_key, None)
            else:
                setattr(self, work_key, getattr(s, work_key))


class BaseSetup:
    """Mixin-class for setups.

    This makes it possible to inherit the most important methods without
    the cumbersome constructor of the ordinary Setup class.

    Maybe this class will be removed in the future, or it could be
    made a proper base class with attributes and so on."""

    orbital_free = False
    hubbard_u = None  # XXX remove me

    def print_info(self, text):
        self.data.print_info(text, self)

    def get_basis_description(self):
        return self.basis.get_description()

    def get_partial_waves_for_atomic_orbitals(self):
        """Get those states phit that represent a real atomic state.

        This typically corresponds to the (truncated) partial waves (PAW) or
        a single-zeta basis."""

        # XXX ugly hack for pseudopotentials:
        if not hasattr(self, 'pseudo_partial_waves_j'):
            return []

        # The zip may cut off part of phit_j if there are more states than
        # projectors.  This should be the correct behaviour for all the
        # currently supported PAW/pseudopotentials.
        partial_waves_j = []
        for n, phit in zips(self.n_j, self.pseudo_partial_waves_j,
                            strict=False):
            if n > 0:
                partial_waves_j.append(phit)
        return partial_waves_j

    def calculate_initial_occupation_numbers(self, magmom, hund, charge,
                                             nspins, f_j=None):
        """If f_j is specified, custom occupation numbers will be used.

        Hund rules disabled if so."""

        nao = self.nao
        f_si = np.zeros((nspins, nao))

        assert (not hund) or f_j is None
        if f_j is None:
            f_j = self.f_j
        f_j = np.array(f_j, float)
        l_j = np.array(self.l_j)

        if hasattr(self, 'data') and hasattr(self.data, 'eps_j'):
            eps_j = np.array(self.data.eps_j)
        else:
            eps_j = np.ones(len(self.n_j))
            # Bound states:
            for j, n in enumerate(self.n_j):
                if n > 0:
                    eps_j[j] = -1.0

        deg_j = 2 * (2 * l_j + 1)

        # Sort after:
        #
        # 1) empty state (f == 0)
        # 2) open shells (d - f)
        # 3) eigenvalues (e)

        states = []
        for j, (f, d, e) in enumerate(zips(f_j, deg_j, eps_j, strict=False)):
            if e < 0.0:
                states.append((f == 0, d - f, e, j))
        states.sort()
        jsorted = [j for _, _, _, j in states]

        # if len(l_j) == 0:
        #     l_j = np.ones(1)

        # distribute the charge to the radial orbitals
        if nspins == 1:
            assert magmom == 0.0
            f_sj = np.array([f_j])
            if not self.orbital_free:
                correct_occ_numbers(f_sj[0], deg_j, jsorted, -charge)
            else:
                # ofdft degeneracy of one orbital is infinite
                f_sj[0] += -charge
        else:
            nval = f_j.sum() - charge
            if np.abs(magmom) > nval:
                raise RuntimeError(
                    'Magnetic moment larger than number ' +
                    f'of valence electrons (|{magmom:g}| > {nval:g})')
            f_sj = 0.5 * np.array([f_j, f_j])
            nup = 0.5 * (nval + magmom)
            ndn = 0.5 * (nval - magmom)
            deg_j //= 2
            correct_occ_numbers(f_sj[0], deg_j, jsorted, nup - f_sj[0].sum())
            correct_occ_numbers(f_sj[1], deg_j, jsorted, ndn - f_sj[1].sum())

        # Projector function indices:
        nj = len(self.n_j)  # or l_j?  Seriously.

        # distribute to the atomic wave functions
        i = 0
        j = 0
        for phit in self.basis_functions_J:
            l = phit.get_angular_momentum_number()

            # Skip functions not in basis set:
            while j < nj and self.l_orb_J[j] != l:
                j += 1
            if j < len(f_j):  # lengths of f_j and l_j may differ
                f = f_j[j]
                f_s = f_sj[:, j]
            else:
                f = 0
                f_s = np.array([0, 0])

            degeneracy = 2 * l + 1

            if hund:
                # Use Hunds rules:
                # assert f == int(f)
                f = int(f)
                f_si[0, i:i + min(f, degeneracy)] = 1.0      # spin up
                f_si[1, i:i + max(f - degeneracy, 0)] = 1.0  # spin down
                if f < degeneracy:
                    magmom -= f
                else:
                    magmom -= 2 * degeneracy - f
            else:
                for s in range(nspins):
                    f_si[s, i:i + degeneracy] = f_s[s] / degeneracy

            i += degeneracy
            j += 1

        if hund and magmom != 0:
            raise WrongMagmomForHundsRuleError(
                f'Bad magnetic moment {magmom:g} for {self.symbol} atom!')
        assert i == nao

        # print('fsi=', f_si)
        return f_si

    def get_hunds_rule_moment(self, charge=0):
        for M in range(10):
            try:
                self.calculate_initial_occupation_numbers(M, True, charge, 2)
            except WrongMagmomForHundsRuleError:
                pass
            else:
                return M
        raise RuntimeError

    def initialize_density_matrix(self, f_si):
        nspins, nao = f_si.shape
        ni = self.ni

        D_sii = np.zeros((nspins, ni, ni))
        D_sp = np.zeros((nspins, ni * (ni + 1) // 2))
        nj = len(self.pt_j)
        j = 0
        i = 0
        ib = 0
        for phit in self.basis_functions_J:
            l = phit.get_angular_momentum_number()
            # Skip functions not in basis set:
            while j < nj and self.l_j[j] != l:
                i += 2 * self.l_j[j] + 1
                j += 1
            if j == nj:
                break

            for m in range(2 * l + 1):
                D_sii[:, i + m, i + m] = f_si[:, ib + m]
            j += 1
            i += 2 * l + 1
            ib += 2 * l + 1
        for s in range(nspins):
            D_sp[s] = pack(D_sii[s])
        return D_sp

    def get_partial_waves(self):
        """Return spline representation of partial waves and densities."""

        l_j = self.l_j

        # cutoffs
        rcut2 = 2 * max(self.rcut_j)
        gcut2 = self.rgd.ceil(rcut2)

        data = self.data

        # Construct splines:
        nc_g = data.nc_g.copy()
        nct_g = data.nct_g.copy()
        tauc_g = data.tauc_g
        tauct_g = data.tauct_g
        nc = self.rgd.spline(nc_g, rcut2, points=1000)
        nct = self.rgd.spline(nct_g, rcut2, points=1000)
        if tauc_g is None:
            tauc_g = np.zeros(nct_g.shape)
            tauct_g = tauc_g
        tauc = self.rgd.spline(tauc_g, rcut2, points=1000)
        tauct = self.rgd.spline(tauct_g, rcut2, points=1000)
        phi_j = []
        phit_j = []
        for j, (phi_g, phit_g) in enumerate(zips(data.phi_jg, data.phit_jg)):
            l = l_j[j]
            phi_g = phi_g.copy()
            phit_g = phit_g.copy()
            phi_g[gcut2:] = phit_g[gcut2:] = 0.0
            phi_j.append(self.rgd.spline(phi_g, rcut2, l, points=100))
            phit_j.append(self.rgd.spline(phit_g, rcut2, l, points=100))
        return phi_j, phit_j, nc, nct, tauc, tauct

    def four_phi_integrals(self):
        """Calculate four-phi integral.

        Calculate the integral over the product of four all electron
        functions in the augmentation sphere, i.e.::

          /
          | d vr  ( phi_i1 phi_i2 phi_i3 phi_i4
          /         - phit_i1 phit_i2 phit_i3 phit_i4 ),

        where phi_i1 is an all electron function and phit_i1 is its
        smooth partner.
        """
        if hasattr(self, 'I4_pp'):
            return self.I4_pp

        # radial grid
        r2dr_g = self.rgd.r_g**2 * self.rgd.dr_g

        phi_jg = self.data.phi_jg
        phit_jg = self.data.phit_jg

        # compute radial parts
        nj = len(self.l_j)
        R_jjjj = np.empty((nj, nj, nj, nj))
        for j1 in range(nj):
            for j2 in range(nj):
                for j3 in range(nj):
                    for j4 in range(nj):
                        R_jjjj[j1, j2, j3, j4] = np.dot(
                            r2dr_g,
                            phi_jg[j1] * phi_jg[j2] *
                            phi_jg[j3] * phi_jg[j4] -
                            phit_jg[j1] * phit_jg[j2] *
                            phit_jg[j3] * phit_jg[j4])

        # prepare for angular parts
        L_i = []
        j_i = []
        for j, l in enumerate(self.l_j):
            for m in range(2 * l + 1):
                L_i.append(l**2 + m)
                j_i.append(j)
        ni = len(L_i)
        # j_i is the list of j values
        # L_i is the list of L (=l**2+m for 0<=m<2*l+1) values
        # https://wiki.fysik.dtu.dk/gpaw/devel/overview.html

        G_LLL = gaunt(max(self.l_j))

        # calculate the integrals
        _np = ni * (ni + 1) // 2  # length for packing
        self.I4_pp = np.empty((_np, _np))
        p1 = 0
        for i1 in range(ni):
            L1 = L_i[i1]
            j1 = j_i[i1]
            for i2 in range(i1, ni):
                L2 = L_i[i2]
                j2 = j_i[i2]
                p2 = 0
                for i3 in range(ni):
                    L3 = L_i[i3]
                    j3 = j_i[i3]
                    for i4 in range(i3, ni):
                        L4 = L_i[i4]
                        j4 = j_i[i4]
                        self.I4_pp[p1, p2] = (np.dot(G_LLL[L1, L2],
                                                     G_LLL[L3, L4]) *
                                              R_jjjj[j1, j2, j3, j4])
                        p2 += 1
                p1 += 1

        # To unpack into I4_iip do:
        # from gpaw.utilities import unpack
        # I4_iip = np.empty((ni, ni, _np)):
        # for p in range(_np):
        #     I4_iip[..., p] = unpack(I4_pp[:, p])

        return self.I4_pp

    def get_default_nbands(self):
        assert len(self.l_orb_J) == len(self.n_j), (self.l_orb_J, self.n_j)
        return sum([2 * l + 1 for (l, n) in zips(self.l_orb_J, self.n_j)
                    if n > 0])

    def calculate_coulomb_corrections(self, wn_lqg, wnt_lqg, wg_lg, wnc_g,
                                      wmct_g):
        """Calculate "Coulomb" energies."""
        # Can we reduce the excessive parameter passing?
        # Seems so ....
        # Added instance variables
        # T_Lqp = self.local_corr.T_Lqp
        # n_qg = self.local_corr.n_qg
        # Delta_lq = self.local_corr.Delta_lq
        # nt_qg = self.local_corr.nt_qg
        # Local variables derived from instance variables
        _np = self.ni * (self.ni + 1) // 2  # change to inst. att.?
        mct_g = self.local_corr.nct_g + self.Delta0 * self.g_lg[0]  # s.a.
        rdr_g = self.local_corr.rgd2.r_g * \
            self.local_corr.rgd2.dr_g  # change to inst. att.?

        A_q = 0.5 * (np.dot(wn_lqg[0], self.local_corr.nc_g) + np.dot(
            self.local_corr.n_qg, wnc_g))
        A_q -= sqrt(4 * pi) * self.Z * np.dot(self.local_corr.n_qg, rdr_g)
        A_q -= 0.5 * (np.dot(wnt_lqg[0], mct_g) +
                      np.dot(self.local_corr.nt_qg, wmct_g))
        A_q -= 0.5 * (np.dot(mct_g, wg_lg[0]) +
                      np.dot(self.g_lg[0], wmct_g)) * \
            self.local_corr.Delta_lq[0]
        M_p = np.dot(A_q, self.local_corr.T_Lqp[0])

        A_lqq = []
        for l in range(2 * self.local_corr.lcut + 1):
            A_qq = 0.5 * np.dot(self.local_corr.n_qg, np.transpose(wn_lqg[l]))
            A_qq -= 0.5 * np.dot(self.local_corr.nt_qg,
                                 np.transpose(wnt_lqg[l]))
            if l <= self.lmax:
                A_qq -= 0.5 * np.outer(self.local_corr.Delta_lq[l],
                                       np.dot(wnt_lqg[l], self.g_lg[l]))
                A_qq -= 0.5 * np.outer(np.dot(self.local_corr.nt_qg,
                                              wg_lg[l]),
                                       self.local_corr.Delta_lq[l])
                A_qq -= 0.5 * np.dot(self.g_lg[l], wg_lg[l]) * \
                    np.outer(self.local_corr.Delta_lq[l],
                             self.local_corr.Delta_lq[l])
            A_lqq.append(A_qq)

        M_pp = np.zeros((_np, _np))
        L = 0
        for l in range(2 * self.local_corr.lcut + 1):
            for m in range(2 * l + 1):  # m?
                M_pp += np.dot(np.transpose(self.local_corr.T_Lqp[L]),
                               np.dot(A_lqq[l], self.local_corr.T_Lqp[L]))
                L += 1

        return M_p, M_pp

    def calculate_integral_potentials(self, func):
        """Calculates a set of potentials using func."""
        wg_lg = [func(self, self.g_lg[l], l)
                 for l in range(self.lmax + 1)]
        wn_lqg = [np.array([func(self, self.local_corr.n_qg[q], l)
                            for q in range(self.local_corr.nq)])
                  for l in range(2 * self.local_corr.lcut + 1)]
        wnt_lqg = [np.array([func(self, self.local_corr.nt_qg[q], l)
                             for q in range(self.local_corr.nq)])
                   for l in range(2 * self.local_corr.lcut + 1)]
        wnc_g = func(self, self.local_corr.nc_g, l=0)
        wnct_g = func(self, self.local_corr.nct_g, l=0)
        wmct_g = wnct_g + self.Delta0 * wg_lg[0]
        return wg_lg, wn_lqg, wnt_lqg, wnc_g, wnct_g, wmct_g

    def calculate_yukawa_interaction(self, gamma):
        """Calculate and return the Yukawa based interaction."""
        if self._Mg_pp is not None and gamma == self._gamma:
            return self._Mg_pp  # Cached

        # Solves the radial screened poisson equation for density n_g
        def Yuk(self, n_g, l):
            """Solve radial screened poisson for density n_g."""
            gamma = self._gamma
            return self.local_corr.rgd2.yukawa(n_g, l, gamma) * \
                self.local_corr.rgd2.r_g * self.local_corr.rgd2.dr_g

        self._gamma = gamma
        self._Mg_pp = self.calculate_vvx_interactions(Yuk)
        return self._Mg_pp

    def calculate_erfc_interaction(self, omega):
        """Calculate and return erfc based valence valence
           exchange interactions."""
        hse = RadialHSE(self.local_corr.rgd2, omega).screened_coulomb_dv

        def erfc_interaction(_, n_g, l):
            return hse(n_g, l)
        return self.calculate_vvx_interactions(erfc_interaction)

    def calculate_vvx_interactions(self, interaction):
        """Calculate valence valence interactions for generic
           interaction."""
        (wg_lg, wn_lqg, wnt_lqg, wnc_g, wnct_g, wmct_g) = \
            self.calculate_integral_potentials(interaction)
        self._Mg_pp = self.calculate_coulomb_corrections(
            wn_lqg, wnt_lqg, wg_lg, wnc_g, wmct_g)[1]
        return self._Mg_pp


class LeanSetup(BaseSetup):
    """Setup class with minimal attribute set.

    A setup-like class must define at least the attributes of this
    class in order to function in a calculation."""
    def __init__(self, s, hubbard_u=None):
        """Copies precisely the necessary attributes of the Setup s."""
        # Hubbard U is poked onto the setup in hacky ways.
        # This needs cleaning.
        self.hubbard_u = hubbard_u

        self.lq = s.lq  # Required for LDA+U I think.
        self.type = s.type  # required for writing to file
        self.fingerprint = s.fingerprint  # also req. for writing
        self.filename = s.filename

        self.symbol = s.symbol
        self.Z = s.Z
        self.Nv = s.Nv
        self.Nc = s.Nc

        self.ni = s.ni
        self.nao = s.nao

        self.pt_j = s.pt_j

        self.pseudo_partial_waves_j = s.pseudo_partial_waves_j
        self.basis_functions_J = s.basis_functions_J

        self.Nct = s.Nct
        self.nct = s.nct

        self.lmax = s.lmax
        self.ghat_l = s.ghat_l
        self.vbar = s.vbar

        self.Delta_pL = s.Delta_pL
        self.Delta0 = s.Delta0

        self.E = s.E
        self.Kc = s.Kc

        self.M = s.M
        self.M_p = s.M_p
        self.M_pp = s.M_pp
        self.M_wpp = s.M_wpp
        self.K_p = s.K_p
        self.MB = s.MB
        self.MB_p = s.MB_p

        self.dO_ii = s.dO_ii

        self.xc_correction = s.xc_correction

        # Required to calculate initial occupations
        self.f_j = s.f_j
        self.n_j = s.n_j
        self.l_j = s.l_j
        self.l_orb_J = s.l_orb_J
        self.nj = len(s.l_j)

        self.data = s.data

        # Below are things which are not really used all that much,
        # i.e. shouldn't generally be necessary.  Maybe we can make a system
        # involving dictionaries for these "optional" parameters

        # Required by print_info
        self.rcutfilter = s.rcutfilter
        self.rcore = s.rcore
        self.basis = s.basis  # we don't need nao if we use this instead

        # XXX figure out better way to store these.
        # Refactoring: We should delete this and use psit_j.  However
        # the code depends on psit_j being the *basis* functions sometimes.
        if hasattr(s, 'pseudo_partial_waves_j'):
            self.pseudo_partial_waves_j = s.pseudo_partial_waves_j
        # Can also get rid of the phit_j splines if need be

        self.N0_p = s.N0_p  # req. by estimate_magnetic_moments
        self.nabla_iiv = s.nabla_iiv  # req. by lrtddft
        self.rxnabla_iiv = s.rxnabla_iiv  # req. by lrtddft and lrtddft2

        # XAS stuff
        self.phicorehole_g = s.phicorehole_g  # should be optional
        if s.phicorehole_g is not None:
            self.A_ci = s.A_ci  # oscillator strengths

        # Required to get all electron density
        self.rgd = s.rgd
        self.rcut_j = s.rcut_j

        self.tauct = s.tauct  # required by TPSS, MGGA

        self.Delta_iiL = s.Delta_iiL  # required with external potential

        self.B_ii = s.B_ii  # required for exact inverse overlap operator
        self.dC_ii = s.dC_ii  # required by time-prop tddft with apply_inverse

        # Required by exx
        self.X_p = s.X_p
        self.X_wp = s.X_wp
        self.ExxC = s.ExxC
        self.ExxC_w = s.ExxC_w

        # Required by yukawa rsf
        self.X_pg = s.X_pg
        self.X_gamma = s.X_gamma

        # Required by electrostatic correction
        self.dEH0 = s.dEH0
        self.dEH_p = s.dEH_p

        # Required by utilities/kspot.py (AllElectronPotential)
        self.g_lg = s.g_lg

        # Probably empty dictionary, required by GLLB
        self.extra_xc_data = s.extra_xc_data

        self.orbital_free = s.orbital_free

        # Stuff required by Yukawa RSF to calculate Mg_pp at runtime
        # the calcualtion of Mg_pp at rt is needed for dscf
        if hasattr(s, 'local_corr'):
            self.local_corr = s.local_corr
        else:
            self.local_corr = LocalCorrectionVar(s)
        self._Mg_pp = None
        self._gamma = 0


class Setup(BaseSetup):
    """Attributes:

    ========== =====================================================
    Name       Description
    ========== =====================================================
    ``Z``      Charge
    ``type``   Type-name of setup (eg. 'paw')
    ``symbol`` Chemical element label (eg. 'Mg')
    ``xcname`` Name of xc
    ``data``   Container class for information on the the atom, eg.
               Nc, Nv, n_j, l_j, f_j, eps_j, rcut_j.
               It defines the radial grid by ng and beta, from which
               r_g = beta * arange(ng) / (ng - arange(ng)).
               It stores pt_jg, phit_jg, phi_jg, vbar_g
    ========== =====================================================


    Attributes for making PAW corrections

    ============= ==========================================================
    Name          Description
    ============= ==========================================================
    ``Delta0``    Constant in compensation charge expansion coeff.
    ``Delta_iiL`` Linear term in compensation charge expansion coeff.
    ``Delta_pL``  Packed version of ``Delta_iiL``.
    ``dO_ii``     Overlap coefficients
    ``B_ii``      Projector function overlaps B_ii = <pt_i | pt_i>
    ``dC_ii``     Inverse overlap coefficients
    ``E``         Reference total energy of atom
    ``M``         Constant correction to Coulomb energy
    ``M_p``       Linear correction to Coulomb energy
    ``M_pp``      2nd order correction to Coulomb energy and Exx energy
    ``M_wpp``     2nd order correction to erfc screened Coulomb energy and Exx
                  energy for given w.
    ``Kc``        Core kinetic energy
    ``K_p``       Linear correction to kinetic energy
    ``ExxC``      Core Exx energy
    ``X_p``       Linear correction to Exx energy
    ``MB``        Constant correction due to vbar potential
    ``MB_p``      Linear correction due to vbar potential
    ``dEH0``      Constant correction due to average electrostatic potential
    ``dEH_p``     Linear correction due to average electrostatic potential
    ``I4_iip``    Correction to integrals over 4 all electron wave functions
    ``Nct``       Analytical integral of the pseudo core density ``nct``
    ============= ==========================================================

    It also has the attribute ``xc_correction`` which is an XCCorrection class
    instance capable of calculating the corrections due to the xc functional.


    Splines:

    ========== ============================================
    Name       Description
    ========== ============================================
    ``pt_j``   Projector functions
    ``phit_j`` Pseudo partial waves
    ``vbar``   vbar potential
    ``nct``    Pseudo core density
    ``ghat_l`` Compensation charge expansion functions
    ``tauct``  Pseudo core kinetic energy density
    ========== ============================================
    """
    def __init__(self, data, xc, lmax=0, basis=None, filter=None):
        self.type = data.name

        if not data.is_compatible(xc):
            raise ValueError('Cannot use %s setup with %s functional' %
                             (data.setupname, xc.get_setup_name()))

        self.symbol = data.symbol
        self.data = data

        self.Nc = data.Nc
        self.Nv = data.Nv
        self.Z = data.Z
        l_j = self.l_j = data.l_j
        self.l_orb_J = data.l_orb_J
        n_j = self.n_j = data.n_j
        self.f_j = data.f_j
        self.eps_j = data.eps_j
        nj = self.nj = len(l_j)
        rcut_j = self.rcut_j = data.rcut_j

        self.ExxC = data.ExxC
        self.ExxC_w = data.ExxC_w
        self.X_p = data.X_p
        self.X_wp = data.X_wp

        self.X_gamma = data.X_gamma
        self.X_pg = data.X_pg

        self.orbital_free = data.orbital_free

        pt_jg = data.pt_jg
        phit_jg = data.phit_jg
        phi_jg = data.phi_jg

        self.fingerprint = data.fingerprint
        self.filename = data.filename

        rgd = self.rgd = data.rgd
        r_g = rgd.r_g
        dr_g = rgd.dr_g

        self.lmax = lmax

        self._Mg_pp = None  # Yukawa based corrections
        self._gamma = 0
        # Attributes for run-time calculation of _Mg_pp
        self.local_corr = LocalCorrectionVar(data)

        rcutmax = max(rcut_j)
        rcut2 = 2 * rcutmax
        gcut2 = rgd.ceil(rcut2)
        self.gcut2 = gcut2

        self.gcutmin = rgd.ceil(min(rcut_j))

        vbar_g = data.vbar_g

        if float(data.version) < 0.7 and data.generator_version < 2:
            # Old-style Fourier-filtered datatsets.
            # Find Fourier-filter cutoff radius:
            gcutfilter = rgd.get_cutoff(pt_jg[0])

        elif filter:
            rc = rcutmax
            vbar_g = vbar_g.copy()
            filter(rgd, rc, vbar_g)

            pt_jg = [pt_g.copy() for pt_g in pt_jg]
            for l, pt_g in zips(l_j, pt_jg):
                filter(rgd, rc, pt_g, l)

            for l in range(max(l_j) + 1):
                J = [j for j, lj in enumerate(l_j) if lj == l]
                A_nn = [[rgd.integrate(phit_jg[j1] * pt_jg[j2]) / 4 / pi
                         for j1 in J] for j2 in J]
                B_nn = np.linalg.inv(A_nn)
                pt_ng = np.dot(B_nn, [pt_jg[j] for j in J])
                for n, j in enumerate(J):
                    pt_jg[j] = pt_ng[n]
            gcutfilter = rgd.get_cutoff(pt_jg[0])
        else:
            gcutfilter = rgd.ceil(max(rcut_j))

        if (vbar_g[gcutfilter:] != 0.0).any():
            gcutfilter = rgd.get_cutoff(vbar_g)
            assert r_g[gcutfilter] < 2.0 * max(rcut_j)

        self.rcutfilter = rcutfilter = r_g[gcutfilter]

        ni = 0
        i = 0
        j = 0
        jlL_i = []
        for l, n in zips(l_j, n_j):
            for m in range(2 * l + 1):
                jlL_i.append((j, l, l**2 + m))
                i += 1
            j += 1
        ni = i
        self.ni = ni

        _np = ni * (ni + 1) // 2
        self.local_corr.nq = nj * (nj + 1) // 2

        lcut = max(l_j)
        if 2 * lcut < lmax:
            lcut = (lmax + 1) // 2
        self.local_corr.lcut = lcut

        self.B_ii = self.calculate_projector_overlaps(pt_jg)

        self.fcorehole = data.fcorehole
        self.lcorehole = data.lcorehole
        if data.phicorehole_g is not None:
            if self.lcorehole == 0:
                self.calculate_oscillator_strengths(phi_jg)
            else:
                self.A_ci = None

        # Construct splines:
        self.vbar = rgd.spline(vbar_g, rcutfilter)

        rcore, nc_g, nct_g, nct = self.construct_core_densities(data)
        self.rcore = rcore
        self.nct = nct

        # Construct splines for core kinetic energy density:
        tauct_g = data.tauct_g
        if tauct_g is not None:
            self.tauct = rgd.spline(tauct_g, self.rcore)
        else:
            self.tauct = None

        self.pt_j = self.create_projectors(pt_jg, rcutfilter)

        partial_waves = self.create_basis_functions(phit_jg, rcut2, gcut2)
        self.pseudo_partial_waves_j = partial_waves.tosplines()

        if basis is None:
            basis = partial_waves
            basis_functions_J = self.pseudo_partial_waves_j
        else:
            basis_functions_J = basis.tosplines()

        self.basis_functions_J = basis_functions_J
        self.basis = basis

        self.nao = 0
        for bf in self.basis_functions_J:
            l = bf.get_angular_momentum_number()
            self.nao += 2 * l + 1

        rgd2 = self.local_corr.rgd2 = rgd.new(gcut2)
        r_g = rgd2.r_g
        dr_g = rgd2.dr_g
        phi_jg = np.array([phi_g[:gcut2].copy() for phi_g in phi_jg])
        phit_jg = np.array([phit_g[:gcut2].copy() for phit_g in phit_jg])
        self.local_corr.nc_g = nc_g = nc_g[:gcut2].copy()
        self.local_corr.nct_g = nct_g = nct_g[:gcut2].copy()
        vbar_g = vbar_g[:gcut2].copy()

        extra_xc_data = dict(data.extra_xc_data)
        # Cut down the GLLB related extra data
        for key, item in extra_xc_data.items():
            if len(item) == rgd.N:
                extra_xc_data[key] = item[:gcut2].copy()
        self.extra_xc_data = extra_xc_data

        self.phicorehole_g = data.phicorehole_g
        if self.phicorehole_g is not None:
            self.phicorehole_g = self.phicorehole_g[:gcut2].copy()

        self.local_corr.T_Lqp = self.calculate_T_Lqp(lcut, _np, nj, jlL_i)
        #  set the attributes directly?
        (self.g_lg, self.local_corr.n_qg, self.local_corr.nt_qg,
         self.local_corr.Delta_lq, self.Lmax, self.Delta_pL, self.Delta0,
         self.N0_p) = self.get_compensation_charges(phi_jg, phit_jg, _np,
                                                    self.local_corr.T_Lqp)

        # Solves the radial poisson equation for density n_g
        def H(self, n_g, l):
            return rgd2.poisson(n_g, l) * r_g * dr_g

        (wg_lg, wn_lqg, wnt_lqg, wnc_g, wnct_g, wmct_g) = \
            self.calculate_integral_potentials(H)
        self.wg_lg = wg_lg

        rdr_g = r_g * dr_g
        dv_g = r_g * rdr_g
        A = 0.5 * np.dot(nc_g, wnc_g)
        A -= sqrt(4 * pi) * self.Z * np.dot(rdr_g, nc_g)
        mct_g = nct_g + self.Delta0 * self.g_lg[0]
        # wmct_g = wnct_g + self.Delta0 * wg_lg[0]
        A -= 0.5 * np.dot(mct_g, wmct_g)
        self.M = A
        self.MB = -np.dot(dv_g * nct_g, vbar_g)

        AB_q = -np.dot(self.local_corr.nt_qg, dv_g * vbar_g)
        self.MB_p = np.dot(AB_q, self.local_corr.T_Lqp[0])

        # Correction for average electrostatic potential:
        #
        #   dEH = dEH0 + dot(D_p, dEH_p)
        #
        self.dEH0 = sqrt(4 * pi) * (wnc_g - wmct_g -
                                    sqrt(4 * pi) * self.Z * r_g * dr_g).sum()
        dEh_q = (wn_lqg[0].sum(1) - wnt_lqg[0].sum(1) -
                 self.local_corr.Delta_lq[0] * wg_lg[0].sum())
        self.dEH_p = np.dot(dEh_q, self.local_corr.T_Lqp[0]) * sqrt(4 * pi)

        M_p, M_pp = self.calculate_coulomb_corrections(wn_lqg, wnt_lqg,
                                                       wg_lg, wnc_g, wmct_g)
        self.M_p = M_p
        self.M_pp = M_pp

        self.M_wpp = {}
        for omega in self.ExxC_w:
            self.M_wpp[omega] = self.calculate_erfc_interaction(omega)

        if xc.type == 'GLLB':
            if 'core_f' in self.extra_xc_data:
                self.wnt_lqg = wnt_lqg
                self.wn_lqg = wn_lqg
                self.fc_j = self.extra_xc_data['core_f']
                self.lc_j = self.extra_xc_data['core_l']
                self.njcore = len(self.lc_j)
                if self.njcore > 0:
                    self.uc_jg = self.extra_xc_data['core_states'].reshape(
                        (self.njcore, -1))
                    self.uc_jg = self.uc_jg[:, :gcut2]
                self.phi_jg = phi_jg

        self.Kc = data.e_kinetic_core - data.e_kinetic
        self.M -= data.e_electrostatic
        self.E = data.e_total

        Delta0_ii = unpack(self.Delta_pL[:, 0].copy())
        self.dO_ii = data.get_overlap_correction(Delta0_ii)
        self.dC_ii = self.get_inverse_overlap_coefficients(self.B_ii,
                                                           self.dO_ii)

        self.Delta_iiL = np.zeros((ni, ni, self.Lmax))
        for L in range(self.Lmax):
            self.Delta_iiL[:, :, L] = unpack(self.Delta_pL[:, L].copy())

        self.Nct = data.get_smooth_core_density_integral(self.Delta0)
        self.K_p = data.get_linear_kinetic_correction(self.local_corr.T_Lqp[0])

        self.ghat_l = [rgd2.spline(g_g, rcut2, l, 50)
                       for l, g_g in enumerate(self.g_lg)]

        self.xc_correction = data.get_xc_correction(rgd2, xc, gcut2, lcut)
        self.nabla_iiv = self.get_derivative_integrals(rgd2, phi_jg, phit_jg)
        self.rxnabla_iiv = self.get_magnetic_integrals(rgd2, phi_jg, phit_jg)

    def create_projectors(self, pt_jg, rcut):
        pt_j = []
        for j, pt_g in enumerate(pt_jg):
            l = self.l_j[j]
            pt_j.append(self.rgd.spline(pt_g, rcut, l))
        return pt_j

    def get_inverse_overlap_coefficients(self, B_ii, dO_ii):
        ni = len(B_ii)
        xO_ii = np.dot(B_ii, dO_ii)
        return -np.dot(dO_ii, np.linalg.inv(np.identity(ni) + xO_ii))

    def calculate_T_Lqp(self, lcut, _np, nj, jlL_i):
        Lcut = (2 * lcut + 1)**2
        G_LLL = gaunt(max(self.l_j))[:, :, :Lcut]
        LGcut = G_LLL.shape[2]
        T_Lqp = np.zeros((Lcut, self.local_corr.nq, _np))
        p = 0
        i1 = 0
        for j1, l1, L1 in jlL_i:
            for j2, l2, L2 in jlL_i[i1:]:
                if j1 < j2:
                    q = j2 + j1 * nj - j1 * (j1 + 1) // 2
                else:
                    q = j1 + j2 * nj - j2 * (j2 + 1) // 2
                T_Lqp[:LGcut, q, p] = G_LLL[L1, L2]
                p += 1
            i1 += 1
        return T_Lqp

    def calculate_projector_overlaps(self, pt_jg):
        """Compute projector function overlaps B_ii = <pt_i | pt_i>."""
        nj = len(pt_jg)
        B_jj = np.zeros((nj, nj))
        for j1, pt1_g in enumerate(pt_jg):
            for j2, pt2_g in enumerate(pt_jg):
                B_jj[j1, j2] = self.rgd.integrate(pt1_g * pt2_g) / (4 * pi)
        B_ii = np.zeros((self.ni, self.ni))
        i1 = 0
        for j1, l1 in enumerate(self.l_j):
            for m1 in range(2 * l1 + 1):
                i2 = 0
                for j2, l2 in enumerate(self.l_j):
                    for m2 in range(2 * l2 + 1):
                        if l1 == l2 and m1 == m2:
                            B_ii[i1, i2] = B_jj[j1, j2]
                        i2 += 1
                i1 += 1
        return B_ii

    def get_compensation_charges(self, phi_jg, phit_jg, _np, T_Lqp):
        lmax = self.lmax
        gcut2 = self.gcut2
        nq = self.local_corr.nq

        g_lg = self.data.create_compensation_charge_functions(lmax)

        n_qg = np.zeros((nq, gcut2))
        nt_qg = np.zeros((nq, gcut2))
        q = 0  # q: common index for j1, j2
        for j1 in range(self.nj):
            for j2 in range(j1, self.nj):
                n_qg[q] = phi_jg[j1] * phi_jg[j2]
                nt_qg[q] = phit_jg[j1] * phit_jg[j2]
                q += 1

        gcutmin = self.gcutmin
        r_g = self.local_corr.rgd2.r_g
        dr_g = self.local_corr.rgd2.dr_g
        self.lq = np.dot(n_qg[:, :gcutmin], r_g[:gcutmin]**2 * dr_g[:gcutmin])

        Delta_lq = np.zeros((lmax + 1, nq))
        for l in range(lmax + 1):
            Delta_lq[l] = np.dot(n_qg - nt_qg, r_g**(2 + l) * dr_g)

        Lmax = (lmax + 1)**2
        Delta_pL = np.zeros((_np, Lmax))
        for l in range(lmax + 1):
            L = l**2
            for m in range(2 * l + 1):
                delta_p = np.dot(Delta_lq[l], T_Lqp[L + m])
                Delta_pL[:, L + m] = delta_p

        Delta0 = np.dot(self.local_corr.nc_g - self.local_corr.nct_g,
                        r_g**2 * dr_g) - self.Z / sqrt(4 * pi)

        # Electron density inside augmentation sphere.  Used for estimating
        # atomic magnetic moment:
        rcutmax = max(self.rcut_j)
        gcutmax = self.rgd.round(rcutmax)
        N0_q = np.dot(n_qg[:, :gcutmax], (r_g**2 * dr_g)[:gcutmax])
        N0_p = np.dot(N0_q, T_Lqp[0]) * sqrt(4 * pi)

        return (g_lg[:, :gcut2].copy(), n_qg, nt_qg,
                Delta_lq, Lmax, Delta_pL, Delta0, N0_p)

    def get_derivative_integrals(self, rgd, phi_jg, phit_jg):
        """Calculate PAW-correction matrix elements of nabla.

        ::

          /  _       _  d       _     ~   _  d   ~   _
          | dr [phi (r) -- phi (r) - phi (r) -- phi (r)]
          /        1    dx    2         1    dx    2

        and similar for y and z."""
        # lmax needs to be at least 1 for evaluating
        # the Gaunt coefficients from derivatives
        lmax = max(1, max(self.l_j))
        G_LLL = gaunt(lmax)
        Y_LLv = nabla(lmax)

        r_g = rgd.r_g
        dr_g = rgd.dr_g
        nabla_iiv = np.empty((self.ni, self.ni, 3))
        if debug:
            nabla_iiv[:] = np.nan
        i1 = 0
        for j1 in range(self.nj):
            l1 = self.l_j[j1]
            nm1 = 2 * l1 + 1
            i2 = 0
            for j2 in range(self.nj):
                l2 = self.l_j[j2]
                nm2 = 2 * l2 + 1
                f1f2or = np.dot(phi_jg[j1] * phi_jg[j2] -
                                phit_jg[j1] * phit_jg[j2], r_g * dr_g)
                dphidr_g = np.empty_like(phi_jg[j2])
                rgd.derivative_spline(phi_jg[j2], dphidr_g)
                dphitdr_g = np.empty_like(phit_jg[j2])
                rgd.derivative_spline(phit_jg[j2], dphitdr_g)
                f1df2dr = np.dot(phi_jg[j1] * dphidr_g -
                                 phit_jg[j1] * dphitdr_g, r_g**2 * dr_g)
                for v in range(3):
                    Lv = 1 + (v + 2) % 3
                    G_12 = G_LLL[Lv, l1**2:l1**2 + nm1, l2**2:l2**2 + nm2]
                    Y_12 = Y_LLv[l1**2:l1**2 + nm1, l2**2:l2**2 + nm2, v]
                    nabla_iiv[i1:i1 + nm1, i2:i2 + nm2, v] = (
                        sqrt(4 * pi / 3) * (f1df2dr - l2 * f1f2or) * G_12
                        + f1f2or * Y_12)
                i2 += nm2
            i1 += nm1
        if debug:
            assert not np.any(np.isnan(nabla_iiv))
        return nabla_iiv

    def get_magnetic_integrals(self, rgd, phi_jg, phit_jg):
        """Calculate PAW-correction matrix elements of r x nabla.

        ::

          /  _       _          _     ~   _      ~   _
          | dr [phi (r) O  phi (r) - phi (r) O  phi (r)]
          /        1     x    2         1     x    2

                       d      d
          where O  = y -- - z --
                 x     dz     dy

        and similar for y and z."""
        # lmax needs to be at least 1 for evaluating
        # the Gaunt coefficients from derivatives
        lmax = max(1, max(self.l_j))
        G_LLL = gaunt(lmax)
        Y_LLv = nabla(2 * lmax)

        r_g = rgd.r_g
        dr_g = rgd.dr_g
        rxnabla_iiv = np.empty((self.ni, self.ni, 3))
        if debug:
            rxnabla_iiv[:] = np.nan
        i1 = 0
        for j1 in range(self.nj):
            l1 = self.l_j[j1]
            nm1 = 2 * l1 + 1
            i2 = 0
            for j2 in range(self.nj):
                l2 = self.l_j[j2]
                nm2 = 2 * l2 + 1
                f1f2or = np.dot(phi_jg[j1] * phi_jg[j2] -
                                phit_jg[j1] * phit_jg[j2], r_g**2 * dr_g)
                for v in range(3):
                    v1 = (v + 1) % 3
                    v2 = (v + 2) % 3
                    Lv1 = 1 + (v1 + 2) % 3
                    Lv2 = 1 + (v2 + 2) % 3
                    # term from radial wfs does not contribute
                    # term from spherical harmonics derivatives
                    G_12 = np.zeros((nm1, nm2))
                    G_12 += np.dot(G_LLL[Lv1, l1**2:l1**2 + nm1, :],
                                   Y_LLv[:, l2**2:l2**2 + nm2, v2])
                    G_12 -= np.dot(G_LLL[Lv2, l1**2:l1**2 + nm1, :],
                                   Y_LLv[:, l2**2:l2**2 + nm2, v1])
                    rxnabla_iiv[i1:i1 + nm1, i2:i2 + nm2, v] = (
                        sqrt(4 * pi / 3) * f1f2or * G_12)
                i2 += nm2
            i1 += nm1
        if debug:
            assert not np.any(np.isnan(rxnabla_iiv))
        return rxnabla_iiv

    def construct_core_densities(self, setupdata):
        rcore = self.data.find_core_density_cutoff(setupdata.nc_g)
        nct = self.rgd.spline(setupdata.nct_g, rcore)
        return rcore, setupdata.nc_g, setupdata.nct_g, nct

    def create_basis_functions(self, phit_jg, rcut2, gcut2):
        # Cutoff for atomic orbitals used for initial guess:
        rcut3 = 8.0  # XXXXX Should depend on the size of the atom!
        gcut3 = self.rgd.ceil(rcut3)

        # We cut off the wave functions smoothly at rcut3 by the
        # following replacement:
        #
        #            /
        #           | f(r),                                   r < rcut2
        #  f(r) <- <  f(r) - a(r) f(rcut3) - b(r) f'(rcut3),  rcut2 < r < rcut3
        #           | 0,                                      r > rcut3
        #            \
        #
        # where a(r) and b(r) are 4. order polynomials:
        #
        #  a(rcut2) = 0,  a'(rcut2) = 0,  a''(rcut2) = 0,
        #  a(rcut3) = 1, a'(rcut3) = 0
        #  b(rcut2) = 0, b'(rcut2) = 0, b''(rcut2) = 0,
        #  b(rcut3) = 0, b'(rcut3) = 1
        #
        r_g = self.rgd.r_g
        x = (r_g[gcut2:gcut3] - rcut2) / (rcut3 - rcut2)
        a_g = 4 * x**3 * (1 - 0.75 * x)
        b_g = x**3 * (x - 1) * (rcut3 - rcut2)

        class PartialWaveBasis(Basis):  # yuckkk
            def __init__(self, symbol, phit_J):
                Basis.__init__(self, symbol, 'partial-waves', readxml=False)
                self._basis_functions_J = phit_J

            def tosplines(self):
                return self._basis_functions_J

            def get_description(self):
                template = 'Using partial waves for %s as LCAO basis'
                string = template % self.symbol
                return string

        basis_functions_J = []
        for j, phit_g in enumerate(phit_jg):
            if self.n_j[j] > 0:
                l = self.l_j[j]
                phit_g = phit_g.copy()
                phit = phit_g[gcut3]
                dphitdr = ((phit - phit_g[gcut3 - 1]) /
                           (r_g[gcut3] - r_g[gcut3 - 1]))
                phit_g[gcut2:gcut3] -= phit * a_g + dphitdr * b_g
                phit_g[gcut3:] = 0.0
                basis_function = self.rgd.spline(phit_g, rcut3, l, points=100)
                basis_functions_J.append(basis_function)
        basis = PartialWaveBasis(self.symbol, basis_functions_J)
        return basis

    def calculate_oscillator_strengths(self, phi_jg):
        # XXX implement oscillator strengths for lcorehole != 0
        assert self.lcorehole == 0
        self.A_ci = np.zeros((3, self.ni))
        nj = len(phi_jg)
        i = 0
        for j in range(nj):
            l = self.l_j[j]
            if l == 1:
                a = self.rgd.integrate(phi_jg[j] * self.data.phicorehole_g,
                                       n=1) / (4 * pi)

                for m in range(3):
                    c = (m + 1) % 3
                    self.A_ci[c, i] = a
                    i += 1
            else:
                i += 2 * l + 1
        assert i == self.ni


class Setups(list):
    """Collection of Setup objects. One for each distinct atom.

    Non-distinct atoms are those with the same atomic number, setup, and basis.

    Class attributes:

    ``nvalence``    Number of valence electrons.
    ``nao``         Number of atomic orbitals.
    ``Eref``        Reference energy.
    ``core_charge`` Core hole charge.
    """

    def __init__(self, Z_a, setup_types, basis_sets, xc, *,
                 filter=None, world=None):
        list.__init__(self)
        symbols = [chemical_symbols[Z] for Z in Z_a]
        type_a = types2atomtypes(symbols, setup_types, default='paw')
        basis_a = types2atomtypes(symbols, basis_sets, default=None)

        for a, _type in enumerate(type_a):
            # Make basis files correspond to setup files.
            #
            # If the setup has a name (i.e. non-default _type), then
            # prepend that name to the basis name.
            #
            # Typically people might specify '11' as the setup but just
            # 'dzp' for the basis set.  Here we adjust to
            # obtain, say, '11.dzp' which loads the correct basis set.
            #
            # There will be no way to obtain the original 'dzp' with
            # a custom-named setup except by loading directly from
            # BasisData.
            #
            # Due to the "szp(dzp)" syntax this is complicated!
            # The name has to go as "szp(name.dzp)".
            basis = basis_a[a]
            if isinstance(basis, str):
                if isinstance(_type, str):
                    setupname = _type
                else:
                    setupname = _type.name  # _type is an object like SetupData
                # Drop DFT+U specification from type string if it is there:
                if hasattr(setupname, 'swapcase'):
                    setupname = setupname.split(':')[0]

                # Basis names inherit setup names except default setups
                # and ghost atoms.
                if setupname != 'paw' and setupname != 'ghost':
                    if setupname:
                        if '(' in basis:
                            reduced, name = basis.split('(')
                            assert name.endswith(')')
                            name = name[:-1]
                            fullname = f'{reduced}({setupname}.{name})'
                        else:
                            fullname = f'{setupname}.{basis_a[a]}'
                        basis_a[a] = fullname

        # Construct necessary PAW-setup objects:
        self.setups = {}
        natoms = {}
        Mcumulative = 0
        self.M_a = []
        self.id_a = list(zips(Z_a, type_a, basis_a))
        for id in self.id_a:
            setup = self.setups.get(id)
            if setup is None:
                Z, type, basis = id
                symbol = chemical_symbols[Z]
                setupdata = None
                if not isinstance(type, str):
                    setupdata = type
                # Basis may be None (meaning that the setup decides), a string
                # (meaning we load the basis set now from a file) or an actual
                # pre-created Basis object (meaning we just pass it along)
                if isinstance(basis, str):
                    basis = Basis(symbol, basis, world=world)
                setup = create_setup(symbol, xc, 2, type,
                                     basis, setupdata=setupdata,
                                     filter=filter, world=world)
                self.setups[id] = setup
                natoms[id] = 0
            natoms[id] += 1
            self.append(setup)
            self.M_a.append(Mcumulative)
            Mcumulative += setup.nao

        # Sum up ...
        self.nvalence = 0       # number of valence electrons
        self.nao = 0            # number of atomic orbitals
        self.Eref = 0.0         # reference energy
        self.core_charge = 0.0  # core hole charge
        for id, setup in self.setups.items():
            n = natoms[id]
            self.Eref += n * setup.E
            self.core_charge += n * (setup.Z - setup.Nv - setup.Nc)
            self.nvalence += n * setup.Nv
            self.nao += n * setup.nao

        self.dS = OverlapCorrections(self)

    def __str__(self):
        # Write PAW setup information in order of appearance:
        ids = set()
        s = 'species:\n'
        for id in self.id_a:
            if id in ids:
                continue
            ids.add(id)
            setup = self.setups[id]
            output = StringIO()
            setup.print_info(functools.partial(print, file=output))
            txt = output.getvalue()
            txt += '  # ' + setup.get_basis_description().replace('\n',
                                                                  '\n  # ')
            txt = txt.replace('\n', '\n  ')
            s += '  ' + txt + '\n\n'

        s += f'Reference energy: {self.Eref * units.Hartree:.6f}  # eV\n'
        return s

    def set_symmetry(self, symmetry):
        """Find rotation matrices for spherical harmonics."""
        # XXX It is ugly that we set self.atomrotations from here;
        # it would be better to return it to the caller.
        from gpaw.atomrotations import AtomRotations
        self.atomrotations = AtomRotations(self.setups, self.id_a, symmetry)

    def empty_atomic_matrix(self, ns, atom_partition, dtype=float):
        Dshapes_a = [(ns, setup.ni * (setup.ni + 1) // 2)
                     for setup in self]
        return atom_partition.arraydict(Dshapes_a, dtype)

    def estimate_dedecut(self, ecut):
        from gpaw.utilities.ekin import dekindecut, ekin
        dedecut = 0.0
        e = {}
        for id in self.id_a:
            if id not in e:
                G, de, e0 = ekin(self.setups[id])
                e[id] = -dekindecut(G, de, ecut)
            dedecut += e[id]
        return dedecut

    def basis_indices(self):
        return FunctionIndices([setup.basis_functions_J for setup in self])

    def projector_indices(self):
        return FunctionIndices([setup.pt_j for setup in self])

    def create_pseudo_core_densities(self, domain, positions, atomdist,
                                     xp=np):
        spline_aj = []
        for setup in self:
            if setup.nct is None:
                spline_aj.append([])
            else:
                spline_aj.append([setup.nct])
        return domain.atom_centered_functions(
            spline_aj, positions,
            atomdist=atomdist,
            integral=[setup.Nct for setup in self],
            cut=True, xp=xp)

    def create_pseudo_core_ked(self,
                               domain,
                               positions,
                               atomdist):
        return domain.atom_centered_functions(
            [[setup.tauct] for setup in self],
            positions,
            atomdist=atomdist,
            cut=True)

    def create_local_potentials(self, domain, positions, atomdist, xp=np):
        return domain.atom_centered_functions(
            [[setup.vbar] for setup in self],
            positions,
            atomdist=atomdist,
            xp=xp)

    def create_compensation_charges(self, domain, positions, atomdist,
                                    xp=np):
        return domain.atom_centered_functions(
            [setup.ghat_l for setup in self], positions,
            atomdist=atomdist,
            integral=sqrt(4 * pi),
            xp=xp)

    def get_overlap_corrections(self, atomdist, xp):
        if atomdist is getattr(self, '_atomdist', None):
            return self.dS_aii
        self._atomdist = atomdist
        dS_aii = AtomArraysLayout([setup.dO_ii.shape for setup in self],
                                  atomdist=atomdist).empty()
        for a, dS_ii in dS_aii.items():
            dS_ii[:] = self[a].dO_ii
        self.dS_aii = dS_aii.to_xp(xp)
        return self.dS_aii

    def partial_wave_corrections(self) -> list[list[Spline]]:
        splines: dict[Setup, list[Spline]] = {}
        dphi_aj = []
        for setup in self:
            dphi_j = splines.get(setup)
            if dphi_j is None:
                rcut = max(setup.rcut_j) * 1.1
                gcut = setup.rgd.ceil(rcut)
                dphi_j = []
                for l, phi_g, phit_g in zips(setup.l_j,
                                             setup.data.phi_jg,
                                             setup.data.phit_jg):
                    dphi_g = (phi_g - phit_g)[:gcut]
                    dphi_j.append(setup.rgd.spline(dphi_g, rcut, l,
                                                   points=200))
                splines[setup] = dphi_j
            dphi_aj.append(dphi_j)

        return dphi_aj


class FunctionIndices:
    def __init__(self, f_aj):
        nm_a = [0]
        for f_j in f_aj:
            nm = sum([2 * f.get_angular_momentum_number() + 1 for f in f_j])
            nm_a.append(nm)
        self.M_a = np.cumsum(nm_a)
        self.nm_a = np.array(nm_a[1:])
        self.max = self.M_a[-1]

    def __getitem__(self, a):
        return self.M_a[a], self.M_a[a + 1]


def types2atomtypes(symbols, types, default):
    """Map a types identifier to a list with a type id for each atom.

    types can be a single str, or a dictionary mapping chemical
    symbols and/or atom numbers to a type identifier.
    If both a symbol key and atomnumber key relates to the same atom, then
    the atomnumber key is dominant.

    If types is a dictionary and contains the string 'default', this will
    be used as default type, otherwize input arg ``default`` is used as
    default.
    """
    natoms = len(symbols)
    if isinstance(types, str):
        return [types] * natoms

    # If present, None will map to the default type,
    # else use the input default
    type_a = [types.get('default', default)] * natoms

    # First symbols ...
    for symbol, type in types.items():
        # Types are given either by strings or they are objects that
        # have a 'symbol' attribute (SetupData, Pseudopotential, Basis, etc.).
        assert isinstance(type, str) or hasattr(type, 'symbol')
        if isinstance(symbol, str):
            for a, symbol2 in enumerate(symbols):
                if symbol == symbol2:
                    type_a[a] = type

    # and then atom indices
    for a, type in types.items():
        if isinstance(a, int):
            type_a[a] = type

    return type_a


if __name__ == '__main__':
    print("""\
This is not the setup.py you are looking for!  This setup.py defines a
Setup class used to hold the atomic data needed for a specific atom.
For building the GPAW code you must use the setup.py setuptools script
at the root of the code tree.  Just do "cd .." and you will be at the
right place.""")
    raise SystemExit



import pytest
from ase import Atom, Atoms
from ase.parallel import parprint

from gpaw import GPAW, mpi
from gpaw.test import equal
from gpaw.lrtddft import LrTDDFT
from gpaw.poisson import FDPoissonSolver
from gpaw.pes.dos import DOSPES
from gpaw.pes.tddft import TDDFTPES


@pytest.mark.lrtddft
def test_lrtddft_pes(in_tmp_dir):
    txt = None
    R = 0.7  # approx. experimental bond length
    a = 3.0
    c = 3.0
    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    H2_plus = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                     Atom('H', (a / 2, a / 2, (c + R) / 2))],
                    cell=(a, a, c))

    xc = 'LDA'
    calc = GPAW(mode='fd', gpts=(12, 12, 12), xc=xc, nbands=1,
                poissonsolver=FDPoissonSolver(),
                parallel={'domain': mpi.world.size},
                spinpol=True, txt=txt)
    H2.calc = calc
    e_H2 = H2.get_potential_energy()

    calc_plus = GPAW(mode='fd', gpts=(12, 12, 12), xc=xc, nbands=2,
                     poissonsolver=FDPoissonSolver(),
                     parallel={'domain': mpi.world.size},
                     spinpol=True, txt=txt)
    calc_plus = calc_plus.new(charge=+1)
    H2_plus.calc = calc_plus
    e_H2_plus = H2_plus.get_potential_energy()

    out = 'dospes.dat'
    pes = DOSPES(calc, calc_plus, shift=True)
    pes.save_folded_pes(filename=out, folding=None)
    parprint('DOS:')
    pes.save_folded_pes(filename=None, folding=None)

    # check for correct shift
    VDE = calc_plus.get_potential_energy() - calc.get_potential_energy()
    BE_HOMO = 1.e23
    be_n, f_n = pes.get_energies_and_weights()
    for be, f in zip(be_n, f_n):
        if f > 0.1 and be < BE_HOMO:
            BE_HOMO = be
    equal(BE_HOMO, VDE)

    lr = LrTDDFT(calc_plus, xc=xc)

    out = 'lrpes.dat'
    pes = TDDFTPES(calc, lr)
    pes.save_folded_pes(filename=out, folding='Gauss')
    parprint('Linear response:')
    pes.save_folded_pes(filename=None, folding=None)

    energy_tolerance = 0.001
    equal(e_H2, -3.90059, energy_tolerance)
    equal(e_H2_plus, 10.5659703, energy_tolerance)

    # io
    out = 'lrpes.dat.gz'
    lr.write(out)
    lr = LrTDDFT.read(out)
    lr.calculator = calc_plus

    pes = TDDFTPES(calc, lr)
    parprint('Linear response:')
    pes.save_folded_pes(filename=None, folding=None)



import time
import numpy as np

from gpaw.analyse.observers import Observer


def time_to_seconds(timestr):
    """Convert time to seconds

    Parameters:

    timestr: float or string
        Float in seconds or string in format
        'DD-HH:MM:SS', 'HH:MM:SS', 'MM:SS', or 'SS'.
    """
    try:
        return float(timestr)
    except ValueError:
        pass
    time = 0.0
    d_i = timestr.split('-')
    if len(d_i) > 1:
        assert len(d_i) == 2
        time += int(d_i[0]) * 24 * 60 * 60
        timestr = d_i[1]
    d_i = timestr.split(':')
    mult = 1
    for d in d_i[::-1]:
        time += int(d) * mult
        mult *= 60
    return time


class TimeLimiter(Observer):
    """Class for automatically breaking the loops of GPAW calculation.

    The time estimation is done by a polynomial fit to
    the data `(i, dt)`, where `i` is the iteration index and
    `dt` is the calculation time of that iteration.

    The loop is broken by adjusting paw.maxiter value (or equivalent).
    """

    # Keywords for supported loops
    scf = 'scf'
    tddft = 'tddft'

    def __init__(self, paw, timestart=None, timelimit='10:00',
                 output=None, interval=1):
        """__init__ method

        Parameters:

        paw:
            GPAW calculator
        timestart: float
            The start time defining the "zero time".
            Format: as given by time.time().
        timelimit: float or string
            The allowed run time counted from `timestart`.
            Format: any supported by function `time_to_seconds()`.
        output: str
            The name of the output file for dumping the time estimates.
        """
        Observer.__init__(self, interval)
        self.timelimit = time_to_seconds(timelimit)
        if timestart is None:
            self.time0 = time.time()
        else:
            self.time0 = timestart
        self.comm = paw.world
        self.output = output
        self.do_output = self.output is not None
        if self.comm.rank == 0 and self.do_output:
            self.outf = open(self.output, 'w')
        self.loop = None
        paw.attach(self, interval, paw)

    def reset(self, loop, order=0, min_updates=5):
        """Reset the time estimation.

        Parameters:

        loop: str
            The keyword of the controlled loop.
        order: int
            The polynomial order of the fit used to estimate
            the run time between each update.
        min_updates: int
            The minimum number of updates until time estimates are given.
        """
        if loop not in [self.scf, self.tddft]:
            raise RuntimeError('Unsupported loop type: {}'.format(loop))
        self.loop = loop
        if self.comm.rank == 0:
            self.order = order
            self.min_updates = max(min_updates, order + 1)
            self.time_t = [time.time()]  # Add the initial time
            self.iteridx_t = []

    def update(self, paw):
        """Update time estimate and break calculation if necessary."""
        # Select the iteration index
        if self.loop is None:
            return
        elif self.loop == self.scf:
            iteridx = paw.scf.niter
        elif self.loop == self.tddft:
            iteridx = paw.niter

        # Update the arrays
        if self.comm.rank == 0:
            self.time_t.append(time.time())
            self.iteridx_t.append(iteridx)
            self.p_k = None

            if self.do_output:
                timediff = self.time_t[-1] - self.time_t[-2]
                line = 'update %12d %12.4f' % (iteridx, timediff)
                self.outf.write('%s\n' % line)
                # self.outf.flush()

        # Check if there is time to do the next iteration
        if not self.has_time(iteridx + self.interval):
            # The calling loop is assumed to do "niter += 1"
            # after calling observers
            paw.log('{}: Breaking the loop '
                    'due to the time limit'.format(self.__class__.__name__))
            if self.loop == self.scf:
                paw.scf.maxiter = iteridx
            elif self.loop == self.tddft:
                paw.maxiter = iteridx

    def eta(self, iteridx):
        """Estimate the time required to calculate the iteration of
           the given index `iteridx`."""
        if self.comm.rank == 0:
            if len(self.iteridx_t) < self.min_updates:
                eta = 0.0
            else:
                if self.p_k is None:
                    iteridx_t = np.array(self.iteridx_t)
                    time_t = np.array(self.time_t)
                    timediff_t = time_t[1:] - time_t[:-1]

                    self.p_k = np.polyfit(iteridx_t,
                                          timediff_t,
                                          self.order)
                if type(iteridx) in (int, float):
                    iteridx = [iteridx]
                iteridx_i = np.array(iteridx)
                eta = max(0.0, np.sum(np.polyval(self.p_k, iteridx_i)))

                if self.do_output:
                    line = 'eta    %12s %12.4f' % (iteridx, eta)
                    self.outf.write('%s\n' % line)
            return eta
        else:
            return None

    def has_time(self, iteridx):
        """Check if there is still time to calculate the iteration of
           the given index `iteridx`."""
        if self.timelimit is None:
            return True
        # Calculate eta on master and broadcast to all ranks
        data_i = np.empty(1, dtype=int)
        if self.comm.rank == 0:
            if len(self.iteridx_t) < self.min_updates:
                data_i[0] = True
            else:
                time_required = self.eta(iteridx)
                time_available = self.timelimit - (time.time() - self.time0)
                data_i[0] = time_required < time_available
        self.comm.broadcast(data_i, 0)
        return bool(data_i[0])

    def __del__(self):
        if self.comm.rank == 0 and self.do_output:
            self.outf.close()



from math import cos, pi, sin

import pytest
from ase import Atom, Atoms

from gpaw import GPAW, Davidson, Mixer, PoissonSolver
from gpaw.test import equal, gen


@pytest.mark.later
def test_corehole_h2o_recursion(in_tmp_dir):
    # Generate setup for oxygen with half a core-hole:
    s = gen('O', name='hch1s', corehole=(1, 0, 0.5))

    a = 5.0
    d = 0.9575
    t = pi / 180 * 104.51
    H2O = Atoms([Atom('O', (0, 0, 0)),
                 Atom('H', (d, 0, 0)),
                 Atom('H', (d * cos(t), d * sin(t), 0))],
                cell=(a, a, a), pbc=False)
    H2O.center()
    calc = GPAW(mode='fd', nbands=10, h=0.2, setups={'O': s},
                eigensolver=Davidson(4),
                mixer=Mixer(0.5),
                xc='oldLDA',
                poissonsolver=PoissonSolver(use_charge_center=True))
    H2O.calc = calc
    e = H2O.get_potential_energy()
    niter = calc.get_number_of_iterations()

    from gpaw.xas import RecursionMethod

    if 1:
        r = RecursionMethod(calc)
        r.run(400)
        r.write('h2o.pckl')
    else:
        r = RecursionMethod(filename='h2o.pckl')

    print(e, niter)
    energy_tolerance = 0.002
    equal(e, -17.980, energy_tolerance)



from pathlib import Path

import pytest
import numpy as np

from ase.build import molecule
from ase.units import Hartree
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.tddft.spectrum import photoabsorption_spectrum as spec_td
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft import photoabsorption_spectrum as spec_lr
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.mpi import world, serial_comm, broadcast

from gpaw.test import only_on_master


pytestmark = [pytest.mark.usefixtures('module_tmp_path')]


@pytest.fixture(scope='module')
@only_on_master(world, broadcast)
def ground_state_calculation():
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='sz(dzp)', mode='lcao', xc='oldLDA',
                convergence={'density': 1e-8},
                communicator=serial_comm,
                symmetry={'point_group': False},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()

    gpw_fpath = Path('gs.gpw').resolve()
    calc.write(gpw_fpath, mode='all')
    return gpw_fpath


@pytest.fixture(scope='module')
def time_propagation_calculation(ground_state_calculation):
    td_calc = LCAOTDDFT(ground_state_calculation, txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick([0, 0, 1e-5])
    td_calc.propagate(30, 150)
    spec_td('dm.dat', 'spec_td.dat',
            e_min=0, e_max=10, width=0.5, delta_e=0.1)
    world.barrier()

    # Scale energy out due to \omega vs \omega_I difference in
    # broadened spectra in RT-TDDFT and LR-TDDFT
    data_ej = np.loadtxt('spec_td.dat')
    spec_e = data_ej[:, 3]
    spec_e[1:] /= data_ej[1:, 0]
    return spec_e


@pytest.fixture(scope='module')
def lrtddft_calculation(ground_state_calculation):
    calc = GPAW(ground_state_calculation, txt=None)
    lr = LrTDDFT(calc, xc='LDA', txt='lr.out')
    lr.diagonalize()
    spec_lr(lr, 'spec_lr.dat',
            e_min=0, e_max=10, width=0.5, delta_e=0.1)
    world.barrier()

    # Scale energy out due to \omega vs \omega_I difference in
    # broadened spectra in RT-TDDFT and LR-TDDFT
    data_ej = np.loadtxt('spec_lr.dat')
    spec_e = data_ej[:, 4]
    spec_e[1:] /= lr[0].get_energy() * Hartree
    return spec_e


@pytest.fixture(scope='module')
def lrtddft2_calculation(ground_state_calculation):
    calc = GPAW(ground_state_calculation, txt='lr2.out')
    lr2 = LrTDDFT2('lr2', calc, fxc='LDA')
    lr2.calculate()
    lr2.get_spectrum('spec_lr2.dat', 0, 10.1, 0.1, width=0.5)
    world.barrier()

    # Scale energy out due to \omega vs \omega_I difference in
    # broadened spectra in RT-TDDFT and LR-TDDFT
    data_ej = np.loadtxt('spec_lr2.dat')
    spec_e = data_ej[:, 5]
    spec_e[1:] /= lr2.lr_transitions.get_transitions()[0][0]
    return spec_e


@pytest.mark.rttddft
def test_lcaotddft_vs_lrtddft(time_propagation_calculation,
                              lrtddft_calculation):
    # One can decrease the tolerance by decreasing the time step
    # and other parameters
    assert (time_propagation_calculation
            == pytest.approx(lrtddft_calculation, abs=1e-2))


@pytest.mark.rttddft
def test_lcaotddft_vs_lrtddft2(time_propagation_calculation,
                               lrtddft2_calculation):
    # One can decrease the tolerance by decreasing the time step
    # and other parameters
    assert (time_propagation_calculation
            == pytest.approx(lrtddft2_calculation, abs=1e-2))


@pytest.mark.rttddft
def test_lrtddft_vs_lrtddft2(lrtddft_calculation,
                             lrtddft2_calculation):
    assert (lrtddft_calculation
            == pytest.approx(lrtddft2_calculation, abs=1e-3))



"""Test of PBLAS Level 2 & 3 : rk, r2k, gemv, gemm.

The test generates random matrices A0, B0, X0, etc. on a
1-by-1 BLACS grid. They are redistributed to a mprocs-by-nprocs
BLACS grid, BLAS operations are performed in parallel, and
results are compared against BLAS.
"""

import pytest
import numpy as np

from gpaw.mpi import world, rank, broadcast_float
from gpaw.test import equal
from gpaw.blacs import BlacsGrid, Redistributor
from gpaw.utilities import compiled_with_sl
from gpaw.utilities.blas import r2k, rk
from gpaw.utilities.scalapack import \
    pblas_simple_gemm, pblas_gemm, \
    pblas_simple_gemv, pblas_gemv, \
    pblas_simple_r2k, pblas_simple_rk, \
    pblas_simple_hemm, pblas_hemm, \
    pblas_simple_symm, pblas_symm
from gpaw.utilities.tools import tri2full

pytestmark = pytest.mark.skipif(not compiled_with_sl(),
                                reason='not compiled with scalapack')

# may need to be be increased if the mprocs-by-nprocs
# BLACS grid becomes larger
tol = 5.0e-13

mnprocs_i = [(1, 1)]
if world.size >= 2:
    mnprocs_i += [(1, 2), (2, 1)]
if world.size >= 4:
    mnprocs_i += [(2, 2)]
if world.size >= 8:
    mnprocs_i += [(2, 4), (4, 2)]


def initialize_random(seed, dtype):
    gen = np.random.Generator(np.random.PCG64(seed))
    if dtype == complex:
        def random(*args):
            return gen.random(*args) + 1.0j * gen.random(*args)
    else:
        def random(*args):
            return gen.random(*args)
    return random


def initialize_alpha_beta(simple, random):
    if simple:
        alpha = 1.0
        beta = 0.0
    else:
        alpha = random()
        beta = random()
    return alpha, beta


def initialize_matrix(grid, M, N, mb, nb, random):
    block_desc = grid.new_descriptor(M, N, mb, nb)
    local_desc = block_desc.as_serial()
    A0 = random(local_desc.shape)
    A0 = np.ascontiguousarray(A0)
    local_desc.checkassert(A0)
    A = local_desc.redistribute(block_desc, A0)
    block_desc.checkassert(A)
    return A0, A, block_desc


def calculate_error(ref_A0, A, block_desc):
    local_desc = block_desc.as_serial()
    A0 = block_desc.redistribute(local_desc, A)
    comm = block_desc.blacsgrid.comm
    if comm.rank == 0:
        err = np.abs(ref_A0 - A0).max()
    else:
        err = np.nan
    err = broadcast_float(err, comm)
    return err


@pytest.mark.parametrize('mprocs, nprocs', mnprocs_i)
@pytest.mark.parametrize('dtype', [float, complex])
def test_pblas_rk_r2k(dtype, mprocs, nprocs,
                      M=160, K=140, seed=42):
    gen = np.random.RandomState(seed)
    grid = BlacsGrid(world, mprocs, nprocs)

    if dtype == complex:
        epsilon = 1.0j
    else:
        epsilon = 0.0

    # Create descriptors for matrices on master:
    globA = grid.new_descriptor(M, K, M, K)
    globD = grid.new_descriptor(M, K, M, K)
    globS = grid.new_descriptor(M, M, M, M)
    globU = grid.new_descriptor(M, M, M, M)

    # print globA.asarray()
    # Populate matrices local to master:
    A0 = gen.rand(*globA.shape) + epsilon * gen.rand(*globA.shape)
    D0 = gen.rand(*globD.shape) + epsilon * gen.rand(*globD.shape)

    # Local result matrices
    S0 = globS.zeros(dtype=dtype)  # zeros needed for rank-updates
    U0 = globU.zeros(dtype=dtype)  # zeros needed for rank-updates

    # Local reference matrix product:
    if rank == 0:
        r2k(1.0, A0, D0, 0.0, S0)
        rk(1.0, A0, 0.0, U0)
    assert globA.check(A0)
    assert globD.check(D0) and globS.check(S0) and globU.check(U0)

    # Create distributed destriptors with various block sizes:
    distA = grid.new_descriptor(M, K, 2, 2)
    distD = grid.new_descriptor(M, K, 2, 3)
    distS = grid.new_descriptor(M, M, 2, 2)
    distU = grid.new_descriptor(M, M, 2, 2)

    # Distributed matrices:
    A = distA.empty(dtype=dtype)
    D = distD.empty(dtype=dtype)
    S = distS.zeros(dtype=dtype)  # zeros needed for rank-updates
    U = distU.zeros(dtype=dtype)  # zeros needed for rank-updates
    Redistributor(world, globA, distA).redistribute(A0, A)
    Redistributor(world, globD, distD).redistribute(D0, D)

    pblas_simple_r2k(distA, distD, distS, A, D, S)
    pblas_simple_rk(distA, distU, A, U)

    # Collect result back on master
    S1 = globS.zeros(dtype=dtype)  # zeros needed for rank-updates
    U1 = globU.zeros(dtype=dtype)  # zeros needed for rank-updates
    Redistributor(world, distS, globS).redistribute(S, S1)
    Redistributor(world, distU, globU).redistribute(U, U1)

    if rank == 0:
        r2k_err = abs(S1 - S0).max()
        rk_err = abs(U1 - U0).max()
        print('r2k err', r2k_err)
        print('rk_err', rk_err)
    else:
        r2k_err = 0.0
        rk_err = 0.0

    # We don't like exceptions on only one cpu
    r2k_err = world.sum(r2k_err)
    rk_err = world.sum(rk_err)

    equal(r2k_err, 0, tol)
    equal(rk_err, 0, tol)


@pytest.mark.parametrize('mprocs, nprocs', mnprocs_i)
@pytest.mark.parametrize('simple', [True, False])
@pytest.mark.parametrize('transa', ['N', 'T', 'C'])
@pytest.mark.parametrize('dtype', [float, complex])
def test_pblas_gemv(dtype, simple, transa, mprocs, nprocs,
                    M=160, N=120, seed=42):
    """Test pblas_simple_gemv, pblas_gemv

    The operation is
    * y <- alpha*A*x + beta*y

    Additional options
    * alpha=1 and beta=0       if simple == True
    """
    random = initialize_random(seed, dtype)
    grid = BlacsGrid(world, mprocs, nprocs)

    # Initialize matrices
    alpha, beta = initialize_alpha_beta(simple, random)
    shapeA = (M, N)
    shapeX = {'N': (N, 1), 'T': (M, 1), 'C': (M, 1)}[transa]
    shapeY = {'N': (M, 1), 'T': (N, 1), 'C': (N, 1)}[transa]
    A0, A, descA = initialize_matrix(grid, *shapeA, 2, 2, random)
    X0, X, descX = initialize_matrix(grid, *shapeX, 4, 1, random)
    Y0, Y, descY = initialize_matrix(grid, *shapeY, 3, 1, random)

    if grid.comm.rank == 0:
        print(A0)

        # Calculate reference with numpy
        op_t = {'N': lambda M: M,
                'T': lambda M: np.transpose(M),
                'C': lambda M: np.conjugate(np.transpose(M))}
        ref_Y0 = alpha * np.dot(op_t[transa](A0), X0) + beta * Y0
    else:
        ref_Y0 = None

    # Calculate with scalapack
    if simple:
        pblas_simple_gemv(descA, descX, descY,
                          A, X, Y,
                          transa=transa)
    else:
        pblas_gemv(alpha, A, X, beta, Y,
                   descA, descX, descY,
                   transa=transa)

    # Check error
    err = calculate_error(ref_Y0, Y, descY)
    assert err < tol


@pytest.mark.parametrize('mprocs, nprocs', mnprocs_i)
@pytest.mark.parametrize('transb', ['N', 'T', 'C'])
@pytest.mark.parametrize('transa', ['N', 'T', 'C'])
@pytest.mark.parametrize('simple', [True, False])
@pytest.mark.parametrize('dtype', [float, complex])
def test_pblas_gemm(dtype, simple, transa, transb, mprocs, nprocs,
                    M=160, N=120, K=140, seed=42):
    """Test pblas_simple_gemm, pblas_gemm

    The operation is
    * C <- alpha*A*B + beta*C

    Additional options
    * alpha=1 and beta=0       if simple == True
    """
    random = initialize_random(seed, dtype)
    grid = BlacsGrid(world, mprocs, nprocs)

    # Initialize matrices
    alpha, beta = initialize_alpha_beta(simple, random)
    shapeA = {'N': (M, K), 'T': (K, M), 'C': (K, M)}[transa]
    shapeB = {'N': (K, N), 'T': (N, K), 'C': (N, K)}[transb]
    shapeC = (M, N)
    A0, A, descA = initialize_matrix(grid, *shapeA, 2, 2, random)
    B0, B, descB = initialize_matrix(grid, *shapeB, 2, 4, random)
    C0, C, descC = initialize_matrix(grid, *shapeC, 3, 2, random)

    if grid.comm.rank == 0:
        print(A0)

        # Calculate reference with numpy
        op_t = {'N': lambda M: M,
                'T': lambda M: np.transpose(M),
                'C': lambda M: np.conjugate(np.transpose(M))}
        ref_C0 = alpha * np.dot(op_t[transa](A0), op_t[transb](B0)) + beta * C0
    else:
        ref_C0 = None

    # Calculate with scalapack
    if simple:
        pblas_simple_gemm(descA, descB, descC,
                          A, B, C,
                          transa=transa, transb=transb)
    else:
        pblas_gemm(alpha, A, B, beta, C,
                   descA, descB, descC,
                   transa=transa, transb=transb)

    # Check error
    err = calculate_error(ref_C0, C, descC)
    assert err < tol


@pytest.mark.parametrize('mprocs, nprocs', mnprocs_i)
@pytest.mark.parametrize('uplo', ['L', 'U'])
@pytest.mark.parametrize('side', ['L', 'R'])
@pytest.mark.parametrize('simple', [True, False])
@pytest.mark.parametrize('hemm', [True, False])
@pytest.mark.parametrize('dtype', [float, complex])
def test_pblas_hemm_symm(dtype, hemm, simple, side, uplo, mprocs, nprocs,
                         M=160, N=120, seed=42):
    """Test pblas_simple_hemm, pblas_simple_symm, pblas_hemm, pblas_symm

    The operation is
    * C <- alpha*A*B + beta*C  if side == 'L'
    * C <- alpha*B*A + beta*C  if side == 'R'

    The computations are done with
    * lower triangular of A    if uplo == 'L'
    * upper triangular of A    if uplo == 'U'

    Additional options
    * A is Hermitian           if hemm == True
    * A is symmetric           if hemm == False
    * alpha=1 and beta=0       if simple == True
    """
    random = initialize_random(seed, dtype)
    grid = BlacsGrid(world, mprocs, nprocs)

    def generate_A_matrix(shape):
        A0 = random(shape)
        if grid.comm.rank == 0:
            if hemm:
                # Hermitian matrix
                A0 = A0 + A0.T.conj()
            else:
                # Symmetric matrix
                A0 = A0 + A0.T

            # Only lower or upper triangular is used, so
            # fill the other triangular with NaN to detect errors
            if uplo == 'L':
                A0 += np.triu(A0 * np.nan, 1)
            else:
                A0 += np.tril(A0 * np.nan, -1)
            A0 = np.ascontiguousarray(A0)
        return A0

    # Initialize matrices
    alpha, beta = initialize_alpha_beta(simple, random)
    shapeA = {'L': (M, M), 'R': (N, N)}[side]
    shapeB = (M, N)
    shapeC = (M, N)
    A0, A, descA = initialize_matrix(grid, *shapeA, 2, 2, generate_A_matrix)
    B0, B, descB = initialize_matrix(grid, *shapeB, 2, 4, random)
    C0, C, descC = initialize_matrix(grid, *shapeC, 3, 2, random)

    if grid.comm.rank == 0:
        print(A0)

        # Calculate reference with numpy
        tri2full(A0, uplo, map=np.conj if hemm else np.positive)
        if side == 'L':
            ref_C0 = alpha * np.dot(A0, B0) + beta * C0
        else:
            ref_C0 = alpha * np.dot(B0, A0) + beta * C0
    else:
        ref_C0 = None

    # Calculate with scalapack
    if simple and hemm:
        pblas_simple_hemm(descA, descB, descC,
                          A, B, C,
                          uplo=uplo, side=side)
    elif hemm:
        pblas_hemm(alpha, A, B, beta, C,
                   descA, descB, descC,
                   uplo=uplo, side=side)
    elif simple:
        pblas_simple_symm(descA, descB, descC,
                          A, B, C,
                          uplo=uplo, side=side)
    else:
        pblas_symm(alpha, A, B, beta, C,
                   descA, descB, descC,
                   uplo=uplo, side=side)

    # Check error
    err = calculate_error(ref_C0, C, descC)
    assert err < tol



from time import ctime

import numpy as np
from ase.units import Ha

from gpaw.response.integrators import Integrand, HilbertTetrahedron, Intraband
from gpaw.response.chi0 import Chi0Calculator
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.chi0_data import Chi0DrudeData
from gpaw.response.frequencies import FrequencyGridDescriptor


class Chi0DrudeCalculator(Chi0Calculator):
    """Class for calculating the plasma frequency contribution to Chi0,
    that is, the contribution from intraband transitions inside of metallic
    bands. This corresponds directly to the dielectric function in the Drude
    model."""

    def __init__(self, *args, **kwargs):
        self.base_ini(*args, **kwargs)
        self.task, self.wd = self.construct_integral_task_and_wd()

    @property
    def nblocks(self):
        # The plasma frequencies aren't distributed in memory
        # NB: There can be a mismatch with self.pair.nblocks, which seems
        # dangerous XXX
        return 1

    def calculate(self, wd, rate, spin='all'):
        """Calculate the Drude dielectric response.

        Parameters
        ----------
        wd : FrequencyDescriptor
            Frequencies to evaluate the reponse function at.
        rate : float
            Plasma frequency decay rate (in eV), corresponding to the
            imaginary part of the complex frequency.
        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin.
        """
        self.print_info(wd, rate)

        # Parse the spin input
        spins = self.get_spins(spin)

        chi0_drude = Chi0DrudeData.from_frequency_descriptor(wd, rate)
        self._calculate(chi0_drude, spins)

        return chi0_drude

    def _calculate(self, chi0_drude: Chi0DrudeData, spins):
        """In-place calculation of the Drude dielectric response function,
        based on the free-space plasma frequency of the intraband transitions.
        """
        # Create a dummy plane-wave descriptor. We need this for the symmetry
        # analysis -> see discussion in gpaw.response.jdos
        qpd = SingleQPWDescriptor.from_q([0., 0., 0.],
                                         ecut=1e-3, gd=self.gs.gd)
        domain, analyzer, prefactor = self.get_integration_domain(qpd, spins)

        # The plasma frequency integral is special in the way that only
        # the spectral part is needed
        integrand = PlasmaFrequencyIntegrand(self, qpd, analyzer)

        # Integrate using temporary array
        tmp_plasmafreq_wvv = np.zeros((1,) + chi0_drude.vv_shape, complex)
        self.integrator.integrate(task=self.task,
                                  domain=domain,  # Integration domain
                                  integrand=integrand,
                                  wd=self.wd,
                                  out_wxx=tmp_plasmafreq_wvv)  # Output array
        tmp_plasmafreq_wvv *= prefactor

        # Store the plasma frequency itself and print it for anyone to use
        plasmafreq_vv = tmp_plasmafreq_wvv[0].copy()
        analyzer.symmetrize_wvv(plasmafreq_vv[np.newaxis])
        chi0_drude.plasmafreq_vv += 4 * np.pi * plasmafreq_vv
        self.context.print('Plasma frequency:', flush=False)
        self.context.print((chi0_drude.plasmafreq_vv**0.5 * Ha).round(2))

        # Calculate the Drude dielectric response function from the
        # free-space plasma frequency
        # χ_D(ω+iη) = ω_p^2 / (ω+iη)^2
        assert chi0_drude.zd.upper_half_plane
        chi0_drude.chi_Zvv += plasmafreq_vv[np.newaxis] \
            / chi0_drude.zd.hz_z[:, np.newaxis, np.newaxis]**2

    def construct_integral_task_and_wd(self):
        if self.integrationmode == 'tetrahedron integration':
            # Calculate intraband transitions at T=0
            fermi_level = self.gs.fermi_level
            wd = FrequencyGridDescriptor([-fermi_level])
            task = HilbertTetrahedron(self.integrator.blockcomm)
        else:
            task = Intraband()

            # We want to pass None for frequency descriptor, but
            # if that goes wrong we'll get TypeError which is unhelpful.
            # This dummy class will give us error messages that allow finding
            # this spot in the code.
            class NotAFrequencyDescriptor:
                pass

            wd = NotAFrequencyDescriptor()
        return task, wd

    def print_info(self, wd, rate):
        isl = ['',
               f'{ctime()}',
               'Calculating drude chi0 with:',
               f'    Number of frequency points:{len(wd)}',
               f'    Plasma frequency decay rate: {rate} eV',
               '',
               self.get_gs_info_string(tab='    ')]
        self.context.print('\n'.join(isl))


class PlasmaFrequencyIntegrand(Integrand):
    def __init__(self, chi0drudecalc, qpd, analyzer):
        self._drude = chi0drudecalc
        self.qpd = qpd
        self.analyzer = analyzer

    def _band_summation(self):
        # Intraband response needs only integrate partially unoccupied bands.
        return self._drude.nocc1, self._drude.nocc2

    def matrix_element(self, k_v, s):
        """NB: In dire need of documentation! XXX."""
        n1, n2 = self._band_summation()
        k_c = np.dot(self.qpd.gd.cell_cv, k_v) / (2 * np.pi)
        kptpair_factory = self._drude.kptpair_factory
        kpt1 = kptpair_factory.get_k_point(s, k_c, n1, n2)
        n_n = range(n1, n2)

        vel_nv = kptpair_factory.pair_calculator().intraband_pair_density(
            kpt1, n_n)

        if self._drude.integrationmode is None:
            f_n = kpt1.f_n
            width = self._drude.gs.get_occupations_width()
            if width > 1e-15:
                dfde_n = - 1. / width * (f_n - f_n**2.0)
            else:
                dfde_n = np.zeros_like(f_n)
            vel_nv *= np.sqrt(-dfde_n[:, np.newaxis])
            weight = np.sqrt(self.analyzer.get_kpoint_weight(k_c) /
                             self.analyzer.how_many_symmetries())
            vel_nv *= weight

        return vel_nv

    def eigenvalues(self, k_v, s):
        """A function that can return the intraband eigenvalues.

        A method describing the integrand of
        the response function which gives an output that
        is compatible with the gpaw k-point integration
        routines."""
        n1, n2 = self._band_summation()
        gs = self._drude.gs
        kd = gs.kd
        k_c = np.dot(self.qpd.gd.cell_cv, k_v) / (2 * np.pi)
        K1 = self._drude.kptpair_factory.find_kpoint(k_c)
        ik = kd.bz2ibz_k[K1]
        kpt1 = gs.kpt_qs[ik][s]
        assert gs.kd.comm.size == 1

        return kpt1.eps_n[n1:n2]



import pytest
from gpaw.sphere.lebedev import run, weight_n, Y_nL, R_nv


@pytest.mark.ci
def test_radial_lebedev(in_tmp_dir):
    weight0_n, Y0_nL, R0_nv = run()
    assert abs(weight0_n - weight_n).sum() < 1e-13
    assert abs(R0_nv - R_nv).sum() < 1e-13
    assert abs(Y0_nL - Y_nL).sum() < 1e-13



class NeedsGD(object):
    """Common base class for classes using a grid descriptor."""

    def __init__(self):
        self.gd = None

    def set_grid_descriptor(self, gd):
        self.gd = gd

    def allocate(self):
        pass

    def estimate_memory(self, mem):
        raise NotImplementedError



import pytest
from ase import Atoms, Atom
from gpaw import GPAW
from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters
from gpaw.test import equal
from gpaw.mpi import world


@pytest.mark.gllb
@pytest.mark.libxc
def test_gllb_ne(in_tmp_dir, add_cwd_to_setup_paths):
    atom = 'Ne'

    for xcname in ['GLLBSC', 'GLLB']:
        if world.rank == 0:
            g = Generator(atom, xcname=xcname, scalarrel=False, nofiles=True)
            g.run(**parameters[atom])
            eps = g.e_j[-1]
        else:
            eps = 0.0
        eps = world.sum(eps)
        world.barrier()

        a = 5
        Ne = Atoms([Atom(atom, (0, 0, 0))],
                   cell=(a, a, a), pbc=False)
        Ne.center()
        calc = GPAW(mode='fd', nbands=7, h=0.25, xc=xcname)
        Ne.calc = calc
        e = Ne.get_potential_energy()
        # Calculate the discontinuity
        homo, lumo = calc.get_homo_lumo()
        response = calc.hamiltonian.xc.response
        dxc_pot = response.calculate_discontinuity_potential(homo, lumo)
        response.calculate_discontinuity(dxc_pot)

        eps3d = calc.wfs.kpt_u[0].eps_n[3]
        # if world.rank == 0:
        equal(eps, eps3d, 1e-3)
        # Correct for small cell +0.14eV (since the test needs to be fast
        # in test suite)
        equal(e + 0.147106041, 0, 5e-2)



import pytest

from gpaw.new.density import atomic_occupation_numbers
from gpaw.setup import create_setup
from gpaw.spinorbit import soc, projected_soc
from gpaw.xc import XC


def test_rad_pot():
    xc = XC('LDA', collinear=False)
    setup = create_setup('Cr', xc=xc)
    mz = 1.0
    f_si = atomic_occupation_numbers(setup, [0, 0, mz], 4)
    print(f_si)
    D_sp = setup.initialize_density_matrix(f_si)
    dv_vii = soc(setup, xc, D_sp)
    # Reversing magmom should give same potential
    D_sp[3] *= -1
    dv_vii -= soc(setup, xc, D_sp)
    assert abs(dv_vii).max() == pytest.approx(0.0, abs=1e-12)


def test_proj_soc():
    xc = XC('LDA', collinear=False)
    setup = create_setup('Cr', xc=xc)
    mz = 1.0
    f_si = atomic_occupation_numbers(setup, [0, 0, mz], 4)
    D_sp = setup.initialize_density_matrix(f_si)
    dv_vii = projected_soc(soc(setup, xc, D_sp),
                           theta=0, phi=0)
    # Reversing normal plane direction should give same potential
    dv_vii -= projected_soc(soc(setup, xc, D_sp),
                            theta=180, phi=0)
    assert abs(dv_vii).max() == pytest.approx(0.0, abs=1e-12)



"""Calculate the Heisenberg exchange constants in Fe and Co using the MFT.
Test with unrealisticly loose parameters to catch if the numerics change.
"""

# General modules
from abc import abstractmethod

import pytest
import numpy as np

# Script modules
from ase.units import Bohr, Ha

from gpaw import GPAW
from gpaw.sphere.integrate import integrate_lebedev

from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.chiks import ChiKSCalculator, smat
from gpaw.response.localft import (LocalFTCalculator, LocalPAWFTCalculator,
                                   add_spin_polarization)
from gpaw.response.mft import (IsotropicExchangeCalculator, AtomicSiteData,
                               StaticSitePairFunction,
                               TwoParticleSiteMagnetizationCalculator,
                               TwoParticleSiteSpinSplittingCalculator)
from gpaw.response.site_kernels import (SphericalSiteKernels,
                                        CylindricalSiteKernels,
                                        ParallelepipedicSiteKernels)
from gpaw.response.heisenberg import (calculate_single_site_magnon_energies,
                                      calculate_fm_magnon_energies)
from gpaw.response.pair_integrator import PairFunctionIntegrator
from gpaw.response.pair_transitions import PairTransitions
from gpaw.response.matrix_elements import (SiteMatrixElementCalculator,
                                           SitePairDensityCalculator,
                                           SitePairSpinSplittingCalculator)
from gpaw.test.conftest import response_band_cutoff
from gpaw.test.response.test_chiks import generate_qrel_q, get_q_c


@pytest.mark.response
def test_Fe_bcc(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    # MFT calculation
    ecut = 50
    # Do the high symmetry points of the bcc lattice
    q_qc = np.array([[0, 0, 0],           # Gamma
                     [0.5, -0.5, 0.5],    # H
                     [0.0, 0.0, 0.5],     # N
                     ])
    # Define site kernels to test
    # Test a single site of spherical and cylindrical geometries
    rc_pa = np.array([[1.0], [1.5], [2.0]])
    hc_pa = np.array([[1.0], [1.5], [2.0]])
    ez_pav = np.array([[[1., 0., 0.]], [[0., 1., 0.]], [[0., 0., 1.]]])

    # ---------- Script ---------- #

    # Extract the ground state fixture
    calc = GPAW(gpw_files['fe_pw'], parallel=dict(domain=1))
    nbands = response_band_cutoff['fe_pw']
    atoms = calc.atoms

    # Set up site kernels with a single site
    positions = atoms.get_positions()
    sitekernels = SphericalSiteKernels(positions, rc_pa)
    sitekernels.append(CylindricalSiteKernels(positions, ez_pav,
                                              rc_pa, hc_pa))
    # Set up a kernel to fill out the entire unit cell
    sitekernels.append(ParallelepipedicSiteKernels(positions,
                                                   [[atoms.get_cell()]]))

    # Initialize the Heisenberg exchange calculator
    gs = ResponseGroundStateAdapter(calc)
    context = ResponseContext()
    chiks_calc = ChiKSCalculator(gs, context,
                                 ecut=ecut, nbands=nbands, gammacentered=True)
    localft_calc = LocalFTCalculator.from_rshe_parameters(gs, context)
    isoexch_calc = IsotropicExchangeCalculator(chiks_calc, localft_calc)

    # Allocate array for the exchange constants
    nq = len(q_qc)
    nsites = sitekernels.nsites
    npartitions = sitekernels.npartitions
    J_qabp = np.empty((nq, nsites, nsites, npartitions), dtype=complex)

    # Calcualate the exchange constant for each q-point
    for q, q_c in enumerate(q_qc):
        J_qabp[q] = isoexch_calc(q_c, sitekernels)

    # Since we only have a single site, reduce the array
    J_qp = J_qabp[:, 0, 0, :]

    # Calculate the magnon energies
    mm = 2.21
    mm_ap = mm * np.ones((1, npartitions))  # Magnetic moments
    mw_qp = calculate_fm_magnon_energies(J_qabp, q_qc, mm_ap)[:, 0, :]

    # Compare results to test values
    test_J_pq = np.array(
        [[2.1907596825086455, 1.172424411323134, 1.6060583789867644],
         [2.612428039019977, 1.2193926800088601, 1.7635196888465006],
         [6.782367391186284, 0.2993922109834177, 1.9346016211386057],
         [1.5764800860123762, 0.8365204592352894, 1.1648584638500161],
         [2.4230224513213234, 1.2179759558303274, 1.6691805687218078],
         [5.35668502504496, 0.3801778545994659, 1.6948968244858478],
         [2.523580017606111, 1.21779750159267, 1.7637120466695273]])
    test_mw_pq = np.array(
        [[0.0, 0.9215703811633589, 0.5291414511510236],
         [0.0, 1.2606654832679791, 0.7682428508357253],
         [0.0, 5.866945864436984, 4.38711834393455],
         [0.0, 0.6696467210652369, 0.3725082553505521],
         [0.0, 1.0905398149239784, 0.682209848506349],
         [0.0, 4.503626398593207, 3.313835475619106],
         [0.0, 1.181703634401304, 0.6876633221145555]])

    # Exchange constants
    assert J_qp.imag == pytest.approx(0.0)
    assert J_qp.T.real == pytest.approx(test_J_pq, rel=2e-3)

    # Magnon energies
    assert mw_qp.T == pytest.approx(test_mw_pq, rel=2e-3)


@pytest.mark.response
def test_Co_hcp(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    # MFT calculation
    ecut = 100
    # Do high symmetry points of the hcp lattice
    q_qc = np.array([[0, 0, 0],              # Gamma
                     [0.5, 0., 0.],          # M
                     [0., 0., 0.5]           # A
                     ])

    # Use spherical site kernels in a radius range which should yield
    # stable results
    rc_pa = np.array([[1.0, 1.0], [1.1, 1.1], [1.2, 1.2]])

    # Unfortunately, the usage of symmetry leads to such extensive repetition
    # of random noise, that one cannot trust individual values of J very well.
    # This is improved when increasing the number of k-points, but the problem
    # never completely vanishes
    J_atol = 1.e-2
    J_rtol = 5.e-2
    # However, derived physical values have an increased error cancellation due
    # to their collective nature.
    mw_rtol = 25e-3  # relative tolerance of absolute results
    mw_ctol = 5.e-2  # relative tolerance on kernel and eta self-consistency

    # ---------- Script ---------- #

    # Extract the ground state fixture
    calc = GPAW(gpw_files['co_pw'], parallel=dict(domain=1))
    nbands = response_band_cutoff['co_pw']
    atoms = calc.get_atoms()

    # Set up spherical site kernels
    positions = atoms.get_positions()
    sitekernels = SphericalSiteKernels(positions, rc_pa)

    # Set up a site kernel to fill out the entire unit cell
    cell_cv = atoms.get_cell()
    cc_v = np.sum(cell_cv, axis=0) / 2.  # Unit cell center
    ucsitekernels = ParallelepipedicSiteKernels([cc_v], [[cell_cv]])

    # Initialize the exchange calculator with and without symmetry
    gs = ResponseGroundStateAdapter(calc)
    context = ResponseContext()
    chiks_calc0 = ChiKSCalculator(gs, context,
                                  disable_point_group=True,
                                  disable_time_reversal=True,
                                  ecut=ecut, nbands=nbands, gammacentered=True)
    localft_calc = LocalPAWFTCalculator(gs, context)
    isoexch_calc0 = IsotropicExchangeCalculator(chiks_calc0, localft_calc)
    chiks_calc1 = ChiKSCalculator(gs, context,
                                  ecut=ecut, nbands=nbands, gammacentered=True)
    isoexch_calc1 = IsotropicExchangeCalculator(chiks_calc1, localft_calc)

    # Allocate array for the spherical site exchange constants
    nq = len(q_qc)
    nsites = sitekernels.nsites
    npartitions = sitekernels.npartitions
    J_qabp = np.empty((nq, nsites, nsites, npartitions), dtype=complex)

    # Allocate array for the unit cell site exchange constants
    Juc_qs = np.empty((nq, 2), dtype=complex)

    # Calcualate the exchange constants for each q-point
    for q, q_c in enumerate(q_qc):
        J_qabp[q] = isoexch_calc0(q_c, sitekernels)
        chiksr_buffer = isoexch_calc0._chiksr
        Juc_qs[q, 0] = isoexch_calc0(q_c, ucsitekernels)[0, 0, 0]
        assert isoexch_calc0._chiksr is chiksr_buffer, \
            'Two subsequent IsotropicExchangeCalculator calls with the same '\
            'q_c, should reuse, not update, the chiks buffer'

        Juc_qs[q, 1] = isoexch_calc1(q_c, ucsitekernels)[0, 0, 0]

    # Calculate the magnon energy
    mom = atoms.get_magnetic_moment()
    mm_ap = mom / 2.0 * np.ones((nsites, npartitions))
    mw_qnp = calculate_fm_magnon_energies(J_qabp, q_qc, mm_ap)
    mw_qnp = np.sort(mw_qnp, axis=1)  # Make sure the eigenvalues are sorted
    mwuc_qs = calculate_single_site_magnon_energies(Juc_qs, q_qc, mom)

    # Compare results to test values
    print(J_qabp[..., 1], mw_qnp[..., 1], mwuc_qs[:, 0])
    test_J_qab = np.array([[[1.23106207 - 0.j, 0.25816335 - 0.j],
                            [0.25816335 + 0.j, 1.23106207 + 0.j]],
                           [[0.88823839 + 0.j, 0.07345416 - 0.04947835j],
                            [0.07345416 + 0.04947835j, 0.88823839 + 0.j]],
                           [[1.09349955 - 0.j, 0.00000010 - 0.01176761j],
                            [0.00000010 + 0.01176761j, 1.09349955 - 0.j]]])
    test_mw_qn = np.array([[0., 0.64793939],
                           [0.64304039, 0.86531921],
                           [0.48182997, 0.51136436]])
    test_mwuc_q = np.array([0., 0.69678659, 0.44825874])

    # Exchange constants
    assert J_qabp[..., 1] == pytest.approx(test_J_qab, abs=J_atol, rel=J_rtol)

    # Magnon energies
    assert np.all(np.abs(mw_qnp[0, 0, :]) < 1.e-8)  # Goldstone theorem
    assert np.allclose(mwuc_qs[0, :], 0.)  # Goldstone
    assert mw_qnp[1:, 0, 1] == pytest.approx(test_mw_qn[1:, 0], rel=mw_rtol)
    assert mw_qnp[:, 1, 1] == pytest.approx(test_mw_qn[:, 1], rel=mw_rtol)
    assert mwuc_qs[1:, 0] == pytest.approx(test_mwuc_q[1:], rel=mw_rtol)

    # Check self-consistency of results
    # We should be in a radius range, where the magnon energies don't change
    assert np.allclose(mw_qnp[1:, 0, ::2],
                       test_mw_qn[1:, 0, np.newaxis], rtol=mw_ctol)
    assert np.allclose(mw_qnp[:, 1, ::2],
                       test_mw_qn[:, 1, np.newaxis], rtol=mw_ctol)
    # Check that symmetry toggle do not change the magnon energies
    assert np.allclose(mwuc_qs[1:, 0], mwuc_qs[1:, 1], rtol=mw_ctol)


@pytest.mark.response
def test_Fe_site_magnetization(gpw_files):
    # Set up ground state adapter
    calc = GPAW(gpw_files['fe_pw'], parallel=dict(domain=1))
    gs = ResponseGroundStateAdapter(calc)

    # Extract valid site radii range
    rmin_a, rmax_a = AtomicSiteData.valid_site_radii_range(gs)
    rmin = rmin_a[0]  # Only one magnetic atom in the unit cell
    rmax = rmax_a[0]
    # We expect rmax to be equal to the nearest neighbour distance
    # subtracted with the augmentation sphere radius. For a bcc lattice,
    # nn_dist = sqrt(3) a / 2:
    augr = gs.get_aug_radii()[0]
    rmax_expected = np.sqrt(3) * 2.867 / 2. - augr * Bohr
    assert abs(rmax - rmax_expected) < 1e-6
    # Test that an error is raised outside the valid range
    with pytest.raises(AssertionError):
        AtomicSiteData(gs, indices=[0],  # Too small radii
                       radii=[np.linspace(rmin * 0.8, rmin, 5)])
    with pytest.raises(AssertionError):
        AtomicSiteData(gs, indices=[0],  # Too large radii
                       radii=[np.linspace(rmax, rmax * 1.2, 5)])
    # Define atomic sites to span the valid range
    rc_r = np.linspace(rmin_a[0], rmax_a[0], 100)
    # Add the radius of the augmentation sphere explicitly
    rc_r = np.append(rc_r, [augr * Bohr])
    atomic_sites = AtomicSiteData(gs, indices=[0], radii=[rc_r])

    # Calculate site magnetization
    magmom_ar = atomic_sites.calculate_magnetic_moments()
    magmom_r = magmom_ar[0]

    # Test that a cutoff at the augmentation sphere radius reproduces
    # the local magnetic moment of the GPAW calculation
    magmom_at_augr = calc.get_atoms().get_magnetic_moments()[0]
    assert abs(magmom_r[-1] - magmom_at_augr) < 1e-2

    # Do a manual calculation of the magnetic moment using the
    # all-electron partial waves
    # Calculate all-electron m(r)
    microsetup = atomic_sites.microsetup_a[0]
    m_ng = np.array([microsetup.rgd.zeros()
                     for n in range(microsetup.Y_nL.shape[0])])
    for n, Y_L in enumerate(microsetup.Y_nL):
        n_sg = np.dot(Y_L, microsetup.n_sLg)
        add_spin_polarization(microsetup.rgd, n_sg, m_ng[n, :])
    # Integrate with varrying radii
    m_g = integrate_lebedev(m_ng)
    ae_magmom_r = np.array([
        microsetup.rgd.integrate_trapz(m_g, rcut=rcut / Bohr)
        for rcut in rc_r])
    # Test that values match approximately inside the augmentation sphere
    inaug_r = rc_r <= augr * Bohr
    assert magmom_r[inaug_r] == pytest.approx(ae_magmom_r[inaug_r], abs=3e-2)

    # import matplotlib.pyplot as plt
    # plt.plot(rc_r[:-1], magmom_r[:-1])
    # plt.plot(rc_r[:-1], ae_magmom_r[:-1], zorder=0)
    # plt.axvline(augr * Bohr, c='0.5', linestyle='--')
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$m$ [$\mu_\mathrm{B}$]')
    # plt.show()


@pytest.mark.response
def test_Co_site_data(gpw_files):
    # Set up ground state adapter
    calc = GPAW(gpw_files['co_pw'], parallel=dict(domain=1))
    gs = ResponseGroundStateAdapter(calc)

    # Extract valid site radii range
    rmin_a, rmax_a = AtomicSiteData.valid_site_radii_range(gs)
    # The valid ranges should be equal due to symmetry
    assert abs(rmin_a[1] - rmin_a[0]) < 1e-8
    assert abs(rmax_a[1] - rmax_a[0]) < 1e-8
    rmin = rmin_a[0]
    rmax = rmax_a[0]
    # We expect rmax to be equal to the nearest neighbour distance
    # subtracted with the augmentation sphere radius. For the hcp-lattice,
    # nn_dist = min(a, sqrt(a^2/3 + c^2/4)):
    augr_a = gs.get_aug_radii()
    assert abs(augr_a[1] - augr_a[0]) < 1e-8
    augr = augr_a[0]
    rmax_expected = min(2.5071, np.sqrt(2.5071**2 / 3 + 4.0695**2 / 4))
    rmax_expected -= augr * Bohr
    assert abs(rmax - rmax_expected) < 1e-6

    # Use radii spanning the entire valid range
    rc_r = np.linspace(rmin, rmax, 101)
    # Add the radius of the augmentation sphere explicitly
    rc_r = np.append(rc_r, [augr * Bohr])
    nr = len(rc_r)
    # Varry the site radii together and independently
    rc1_r = list(rc_r) + list(rc_r) + [augr * Bohr] * nr
    rc2_r = list(rc_r) + [augr * Bohr] * nr + list(rc_r)
    atomic_sites = AtomicSiteData(gs, indices=[0, 1], radii=[rc1_r, rc2_r])

    # Calculate site magnetization
    magmom_ar = atomic_sites.calculate_magnetic_moments()

    # Test that the magnetization inside the augmentation sphere matches
    # the local magnetic moment of the GPAW calculation
    magmom_at_augr_a = calc.get_atoms().get_magnetic_moments()
    assert magmom_ar[:, -1] == pytest.approx(magmom_at_augr_a, abs=2e-2)

    # Test consistency of varrying radii
    assert magmom_ar[0, :nr] == pytest.approx(magmom_ar[1, :nr])
    assert magmom_ar[0, nr:2 * nr] == pytest.approx(magmom_ar[0, :nr])
    assert magmom_ar[0, 2 * nr:] == pytest.approx([magmom_ar[0, -1]] * nr)
    assert magmom_ar[1, nr:2 * nr] == pytest.approx([magmom_ar[1, -1]] * nr)
    assert magmom_ar[1, 2 * nr:] == pytest.approx(magmom_ar[1, :nr])

    # Calculate the atomic spin splitting
    rc_r = rc_r[:-1]
    atomic_sites = AtomicSiteData(gs, indices=[0, 1], radii=[rc_r, rc_r])
    dxc_ar = atomic_sites.calculate_spin_splitting()
    print(dxc_ar[0, ::20])

    # Test that the spin splitting comes out as expected
    assert dxc_ar[0] == pytest.approx(dxc_ar[1])
    assert dxc_ar[0, ::20] == pytest.approx([0.02638351, 1.41476112,
                                             2.49540004, 2.79727200,
                                             2.82727948, 2.83670767], rel=1e-3)

    # import matplotlib.pyplot as plt
    # plt.subplot(1, 2, 1)
    # plt.plot(rc_r, magmom_ar[0, :nr - 1])
    # plt.axvline(augr * Bohr, c='0.5', linestyle='--')
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$m$ [$\mu_\mathrm{B}$]')
    # plt.subplot(1, 2, 2)
    # plt.plot(rc_r, dxc_ar[0])
    # plt.axvline(augr * Bohr, c='0.5', linestyle='--')
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$\Delta_\mathrm{xc}$ [eV]')
    # plt.show()


@pytest.mark.response
@pytest.mark.parametrize('qrel', generate_qrel_q())
def test_Co_site_magnetization_sum_rule(in_tmp_dir, gpw_files, qrel):
    # Set up ground state adapter and atomic site data
    calc = GPAW(gpw_files['co_pw'], parallel=dict(domain=1))
    nbands = response_band_cutoff['co_pw']
    gs = ResponseGroundStateAdapter(calc)
    context = ResponseContext('Co_sum_rule.txt')
    atomic_site_data = get_co_atomic_site_data(gs)
    nblocks = generate_nblocks(context)

    # Get wave vector to test
    q_c = get_q_c('co_pw', qrel)

    # ----- Single-particle site magnetization ----- #
    # Set up calculator and calculate the site magnetization
    simple_site_mag_calc = SingleParticleSiteMagnetizationCalculator(
        gs, context)
    ssite_mag_ar = simple_site_mag_calc(atomic_site_data)

    # Test that the imaginary part vanishes (we use only diagonal pair
    # densities correcsponding to |ψ_nks(r)|^2)
    assert np.allclose(ssite_mag_ar.imag, 0.)
    ssite_mag_ar = ssite_mag_ar.real

    # Test that the results match a conventional calculation
    magmom_ar = atomic_site_data.calculate_magnetic_moments()
    assert ssite_mag_ar == pytest.approx(magmom_ar, rel=5e-3)

    # ----- Two-particle site magnetization ----- #
    # Set up calculator and calculate site magnetization by sum rule
    sum_rule_site_mag_calc = TwoParticleSiteMagnetizationCalculator(
        gs, context, nblocks=nblocks, nbands=nbands)
    site_mag_abr = sum_rule_site_mag_calc(q_c, atomic_site_data)
    context.write_timer()

    # Test that the sum rule site magnetization is a positive-valued diagonal
    # real array
    site_mag_ra = site_mag_abr.diagonal()
    assert np.all(site_mag_ra.real > 0)
    assert np.all(np.abs(site_mag_ra.imag) / site_mag_ra.real < 1e-6)
    site_mag_ra = site_mag_ra.real
    assert np.all(np.abs(np.diagonal(np.fliplr(  # off-diagonal elements
        site_mag_abr))) / site_mag_ra < 5e-2)
    site_mag_ar = site_mag_ra.T
    # Test that the magnetic moments on the two Co atoms are identical
    assert site_mag_ar[0] == pytest.approx(site_mag_ar[1], rel=1e-4)

    # Test that the result more or less matches a conventional calculation at
    # close-packing
    assert np.average(site_mag_ar, axis=0)[-1] == pytest.approx(
        np.average(magmom_ar, axis=0)[-1], rel=5e-2)

    # Test values against reference
    print(np.average(site_mag_ar, axis=0)[::2])
    assert np.average(site_mag_ar, axis=0)[::2] == pytest.approx(
        np.array([3.91823444e-04, 1.45641911e-01, 6.85939109e-01,
                  1.18813171e+00, 1.49761591e+00, 1.58954270e+00]), rel=5e-2)

    # import matplotlib.pyplot as plt
    # rc_r = atomic_site_data.rc_ap[0] * Bohr
    # plt.plot(rc_r, site_mag_ar[0], '-o', mec='k')
    # plt.plot(rc_r, ssite_mag_ar[0], '-o', mec='k', zorder=1)
    # plt.plot(rc_r, magmom_ar[0], '-o', mec='k', zorder=0)
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$m$ [$\mu_\mathrm{B}$]')
    # plt.title(str(q_c))
    # plt.show()


@pytest.mark.response
@pytest.mark.parametrize('qrel', generate_qrel_q())
def test_Co_site_spin_splitting_sum_rule(in_tmp_dir, gpw_files, qrel):
    # Set up ground state adapter and atomic site data
    calc = GPAW(gpw_files['co_pw'], parallel=dict(domain=1))
    nbands = response_band_cutoff['co_pw']
    gs = ResponseGroundStateAdapter(calc)
    context = ResponseContext('Co_sum_rule.txt')
    atomic_site_data = get_co_atomic_site_data(gs)
    nblocks = generate_nblocks(context)

    # Get wave vector to test
    q_c = get_q_c('co_pw', qrel)

    # ----- Single-particle site spin splitting ----- #
    # Set up calculator and calculate the site magnetization
    single_particle_dxc_calc = SingleParticleSiteSpinSplittingCalculator(
        gs, context)
    single_particle_dxc_ar = single_particle_dxc_calc(atomic_site_data)

    # Test that the imaginary part vanishes (we use only diagonal pair
    # spin splitting densities correcsponding to -2W_xc^z(r)|ψ_nks(r)|^2)
    assert np.allclose(single_particle_dxc_ar.imag, 0.)
    single_particle_dxc_ar = single_particle_dxc_ar.real

    # Test that the results match a conventional calculation
    dxc_ar = atomic_site_data.calculate_spin_splitting()
    assert single_particle_dxc_ar == pytest.approx(dxc_ar, rel=5e-3)

    # ----- Two-particle site spin splitting ----- #
    # Set up calculator and calculate site spin splitting by sum rule
    two_particle_dxc_calc = TwoParticleSiteSpinSplittingCalculator(
        gs, context, nblocks=nblocks, nbands=nbands)
    tp_dxc_abr = two_particle_dxc_calc(q_c, atomic_site_data)
    context.write_timer()

    # Test that the two-particle spin splitting is a positive-valued diagonal
    # real array
    tp_dxc_ra = tp_dxc_abr.diagonal()
    assert np.all(tp_dxc_ra.real > 0)
    assert np.all(np.abs(tp_dxc_ra.imag) / tp_dxc_ra.real < 1e-4)
    tp_dxc_ra = tp_dxc_ra.real
    assert np.all(np.abs(np.diagonal(np.fliplr(  # off-diagonal elements
        tp_dxc_abr))) / tp_dxc_ra < 5e-2)
    tp_dxc_ar = tp_dxc_ra.T
    # Test that the spin splitting on the two Co atoms is identical
    assert tp_dxc_ar[0] == pytest.approx(tp_dxc_ar[1], rel=1e-4)

    # Test values against reference
    print(np.average(tp_dxc_ar, axis=0)[::2])
    assert np.average(tp_dxc_ar, axis=0)[::2] == pytest.approx(
        np.array([3.68344584e-04, 3.13780575e-01, 1.35409600e+00,
                  2.14237563e+00, 2.52032513e+00, 2.61406726e+00]), rel=5e-2)

    # import matplotlib.pyplot as plt
    # rc_r = atomic_site_data.rc_ap[0] * Bohr
    # plt.plot(rc_r, tp_dxc_ar[0], '-o', mec='k')
    # plt.plot(rc_r, single_particle_dxc_ar[0], '-o', mec='k', zorder=1)
    # plt.plot(rc_r, dxc_ar[0], '-o', mec='k', zorder=0)
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$\Delta_\mathrm{xc}$ [eV]')
    # plt.title(str(q_c))
    # plt.show()


# ---------- Test functionality ---------- #


def get_co_atomic_site_data(gs):
    # Set up atomic sites
    rmin_a, _ = AtomicSiteData.valid_site_radii_range(gs)
    # Make sure that the two sites do not overlap
    nn_dist = min(2.5071, np.sqrt(2.5071**2 / 3 + 4.0695**2 / 4))
    rc_r = np.linspace(rmin_a[0], nn_dist / 2, 11)
    return AtomicSiteData(gs, indices=[0, 1], radii=[rc_r, rc_r])


def generate_nblocks(context):
    if context.comm.size % 4 == 0:
        nblocks = 4
    elif context.comm.size % 2 == 0:
        nblocks = 2
    else:
        nblocks = 1
    return nblocks


class SingleParticleSiteQuantity(StaticSitePairFunction):
    @property
    def shape(self):
        nsites = self.atomic_site_data.nsites
        npartitions = self.atomic_site_data.npartitions
        return nsites, npartitions


class SingleParticleSiteSumRuleCalculator(PairFunctionIntegrator):
    r"""Calculator for single-particle site sum rules.

    For any site matrix element f^a_(nks,n'k's'), one may define a single-
    particle site sum rule by considering only the diagonal of the matrix
    element:
                 __  __
             1   \   \
    f_a^μ = ‾‾‾  /   /  σ^μ_ss f_nks f^a_(nks,nks)
            N_k  ‾‾  ‾‾
                 k   n,s

    where μ∊{0,z}.
    """

    def __init__(self, gs, context):
        super().__init__(gs, context,
                         disable_point_group=True,
                         disable_time_reversal=True)
        self.matrix_element_calc: SiteMatrixElementCalculator | None = None

    def __call__(self, atomic_site_data):
        self.matrix_element_calc = self.create_matrix_element_calculator(
            atomic_site_data)

        # Set up transitions
        # Loop over bands, which are fully or partially occupied
        nocc2 = self.kptpair_extractor.nocc2
        n_n = list(range(nocc2))
        n_t = np.array(n_n + n_n)
        s_t = np.array([0] * nocc2 + [1] * nocc2)
        transitions = PairTransitions(n1_t=n_t, n2_t=n_t, s1_t=s_t, s2_t=s_t)

        # Set up data object with q=0
        qpd = self.get_pw_descriptor([0., 0., 0.], ecut=1e-3)
        site_quantity = SingleParticleSiteQuantity(qpd, atomic_site_data)

        # Perform actual calculation
        self._integrate(site_quantity, transitions)
        return site_quantity.array

    @abstractmethod
    def create_matrix_element_calculator(
            self, atomic_site_data) -> SiteMatrixElementCalculator:
        """Create the desired site matrix element calculator."""

    def add_integrand(self, kptpair, weight, site_quantity):
        r"""Add the integrand of the outer k-point integral.

        With
                   __
                1  \
        f_a^μ = ‾  /  (...)_k
                V  ‾‾
                   k

        the integrand has to be multiplied with the cell volume V0:
                     __
                     \
        (...)_k = V0 /  σ^μ_ss f_nks f^a_(nks,nks)
                     ‾‾
                     n,s
        """
        # Calculate matrix elements
        site_matrix_element = self.matrix_element_calc(
            kptpair, site_quantity.qpd)
        assert site_matrix_element.tblocks.blockcomm.size == 1
        f_tap = site_matrix_element.get_global_array()

        # Calculate Pauli matrix factors and multiply the occupations
        sigma = self.get_pauli_matrix()
        sigma_t = sigma[kptpair.transitions.s1_t, kptpair.transitions.s2_t]
        f_t = kptpair.get_all(kptpair.ikpt1.f_myt)
        sigmaf_t = sigma_t * f_t

        # Calculate and add integrand
        site_quantity.array[:] += self.gs.volume * weight * np.einsum(
            't, tap -> ap', sigmaf_t, f_tap)

    @abstractmethod
    def get_pauli_matrix(self):
        """Get the desired Pauli matrix σ^μ_ss."""


class SingleParticleSiteMagnetizationCalculator(
        SingleParticleSiteSumRuleCalculator):
    r"""Calculator for the single-particle site magnetization sum rule.

    The site magnetization is calculated from the site pair density:
                 __  __
             1   \   \
    n_a^z = ‾‾‾  /   /  σ^z_ss f_nks n^a_(nks,nks)
            N_k  ‾‾  ‾‾
                 k   n,s
    """
    def get_pauli_matrix(self):
        return smat('z')

    def create_matrix_element_calculator(self, atomic_site_data):
        return SitePairDensityCalculator(self.gs, self.context,
                                         atomic_site_data)


class SingleParticleSiteSpinSplittingCalculator(
        SingleParticleSiteMagnetizationCalculator):
    r"""Calculator for the single-particle site spin splitting sum rule.
                      __  __
                  1   \   \
    Δ^(xc)_a^z = ‾‾‾  /   /  σ^z_ss f_nks Δ^(xc,a)_(nks,nks)
                 N_k  ‾‾  ‾‾
                      k   n,s
    """
    def create_matrix_element_calculator(self, atomic_site_data):
        return SitePairSpinSplittingCalculator(self.gs, self.context,
                                               atomic_site_data,
                                               rshewmin=1e-8)

    def __call__(self, *args):
        dxc_ap = super().__call__(*args)
        return dxc_ap * Ha  # Ha -> eV



from optparse import OptionParser


def build_parser():
    parser = OptionParser(usage='%prog [options] [elements]',
                          version='%prog 0.1')
    parser.add_option('-f', '--xcfunctional', type='string', default='LDA',
                      help='Exchange-Correlation functional ' +
                      '(default value LDA)', metavar='<XC>')
    parser.add_option('-n', '--non-scalar-relativistic', action='store_true',
                      default=False,
                      help='Do *not* do a scalar-relativistic calculation.')
    parser.add_option('-x', '--exact-exchange', action='store_true',
                      default=False,
                      help='Calculate exact exchange integrals.')
    parser.add_option('--gamma', type='float',
                      help='Set gamma for Yukawa dampening.')
    parser.add_option('-r', '--radius', type='string', default=None,
                      help='Cutoff radius or radii (comma separated).',
                      metavar='<rcut>')
    parser.add_option('-v', '--zero-potential', metavar='type,radius',
                      help='Type of zero-potential - type must be either ' +
                      '"poly" or "f".')
    parser.add_option('--filter', metavar='h,x',
                      help='Parameters used for Fourier-filtering and '
                      'projector functions and zero-potential. "h" is '
                      'the cutoff grid-spacing (in Bohr) and "x" is the '
                      'ratio between outer and inner radii.')
    parser.add_option('-l', '--logarithmic-derivatives', action='store_true',
                      help='Calculate logarithmic derivatives.')
    parser.add_option('-a', '--all-electron-only', action='store_true',
                      help='Skip generation of PAW setup.')
    parser.add_option('-e', '--extra-projectors', type='string', default=None,
                      help='Extra projectors. Use ";" to separate s, p and ' +
                      'd channels.  Examples: "0.0,1.0" for two extra ' +
                      's-type. "0.0;1.0" for extra s and p. ";1.0" for ' +
                      'extra p.', metavar='0.0;0.0,1.0;0.0')
    parser.add_option('-c', '--core', type='string', default=None,
                      help='Frozen core.  Examples: "[Ne]", "[Ar]3d".',
                      metavar='<core>')
    parser.add_option('--normconserving', type='string',
                      help='Examples: s, sp.')
    parser.add_option('--core-hole', metavar='state,occ',
                      help='Add core hole. Examples: "1s,0.5", "2p,1".')
    parser.add_option('--configuration', metavar='config',
                      help='Specify non-groundstate configuration. '
                      'Na+ ion: "Ne,3s0", O2- ion: "1s2,2s2,2p6" or ' +
                      '"He,2s2,2p6".')
    parser.add_option('--compensation-charge-radius', metavar='rcut',
                      type=float,
                      help='Cutoff radius for compensation charges.')
    parser.add_option('--name', type='string', metavar='<id>',
                      help='Name to use for setup file: <symbol>.<id>.<xc>.  '
                      'Default name is <symbol>.<xc>.')
    parser.add_option('-w', '--write-files', action='store_true',
                      help='Write wave functions and other things to files.')
    parser.add_option('-p', '--plot', action='store_true',
                      help='Show plot and generate reStructuredText.')
    parser.add_option('-g', '--points-per-node', metavar='<gpernode>',
                      type=int, default=150,
                      help='Number of radial grid points per node.')
    parser.add_option('--empty-states', type='string', default=None,
                      help='Add empty state(s).  Example: 5p.',
                      metavar='<states>')
    parser.add_option('--tw-coefficient', type='float', default=1,
                      help='Sets value of coefficient in Thomas-Fermi ' +
                      'calculations. Default is 1',
                      metavar='<tw-coefficient>')
    parser.add_option('--orbital-free', action='store_true',
                      help='Generates orbital-free Thomas-Fermi setup')
    return parser


def main():
    parser = build_parser()
    opt, args = parser.parse_args()

    import sys

    from gpaw.atom.generator import Generator
    from gpaw.atom.configurations import parameters, tf_parameters
    from gpaw.atom.all_electron import AllElectron
    from gpaw import ConvergenceError

    if args:
        atoms = args
    else:
        atoms = parameters.keys()

    bad_density_warning = """\
    Problem with initial electron density guess!  Try to run the program
    with the '-nw' option (non-scalar-relativistic calculation + write
    density) and then try again without the '-n' option (this will
    generate a good initial guess for the density)."""

    for symbol in atoms:
        scalarrel = not opt.non_scalar_relativistic

        corehole = None
        if opt.core_hole is not None:
            state, occ = opt.core_hole.split(',')
            # Translate corestate string ('1s') to n and l:
            ncorehole = int(state[0])
            lcorehole = 'spdf'.find(state[1])
            fcorehole = float(occ)
            corehole = (ncorehole, lcorehole, fcorehole)

        if opt.all_electron_only:
            a = AllElectron(symbol, opt.xcfunctional, scalarrel, corehole,
                            opt.configuration, not opt.write_files, '-',
                            opt.points_per_node,
                            opt.orbital_free, opt.tw_coefficient)
            try:
                a.run()
            except ConvergenceError:
                print(bad_density_warning, file=sys.stderr)
            continue
        g = Generator(symbol, opt.xcfunctional, scalarrel, corehole,
                      opt.configuration, not opt.write_files, '-',
                      opt.points_per_node, orbital_free=opt.orbital_free,
                      tw_coeff=opt.tw_coefficient)

        if opt.orbital_free:
            p = tf_parameters.get(symbol, {'rcut': 0.9})
        else:
            p = parameters.get(symbol, {})

        if opt.core is not None:
            p['core'] = opt.core

        if opt.radius is not None:
            p['rcut'] = [float(x) for x in opt.radius.split(',')]

        if opt.extra_projectors is not None:
            extra = {}
            if opt.extra_projectors != '':
                for l, x in enumerate(opt.extra_projectors.split(';')):
                    if x != '':
                        extra[l] = [float(y) for y in x.split(',')]
            p['extra'] = extra

        if opt.normconserving is not None:
            p['normconserving'] = opt.normconserving

        if opt.filter is not None:
            p['filter'] = [float(x) for x in opt.filter.split(',')]

        if opt.compensation_charge_radius is not None:
            p['rcutcomp'] = opt.compensation_charge_radius

        if opt.zero_potential is not None:
            vbar = opt.zero_potential.split(',')
            p['vbar'] = (vbar[0], float(vbar[1]))

        if opt.empty_states is not None:
            p['empty_states'] = opt.empty_states

        if opt.gamma is not None:
            p['yukawa_gamma'] = opt.gamma

        try:
            g.run(logderiv=opt.logarithmic_derivatives,
                  exx=opt.exact_exchange, name=opt.name,
                  **p)
        except ConvergenceError:
            print(bad_density_warning, file=sys.stderr)
        except RuntimeError as m:
            if len(m.__str__()) == 0:
                raise
            print(m)

        if opt.plot:
            from gpaw.atom.analyse_setup import analyse
            analyse(g, show=True)



from math import pi, sqrt
import numpy as np
from gpaw.utilities.tools import coordinates
from gpaw.utilities.gauss import Gaussian
from gpaw.grid_descriptor import GridDescriptor
from gpaw.test import equal
from gpaw.poisson import PoissonSolver


def test_gauss_func():
    def norm(a):
        return np.sqrt(np.sum(a.ravel()**2)) / len(a.ravel())

    # Initialize classes
    a = 20  # Size of cell
    N = 48  # Number of grid points
    Nc = (N, N, N)  # Number of grid points along each axis
    gd = GridDescriptor(Nc, (a, a, a), 0)  # Grid-descriptor object
    solver = PoissonSolver(nn=3)  # Numerical poisson solver
    solver.set_grid_descriptor(gd)
    solve = solver.solve
    xyz, r2 = coordinates(gd)  # Matrix with square of the radial coordinate
    print(r2.shape)
    r = np.sqrt(r2)  # Matrix with the values of the radial coordinate
    nH = np.exp(-2 * r) / pi     # Density of the hydrogen atom
    gauss = Gaussian(gd)          # An instance of Gaussian

    # /------------------------------------------------\
    # | Check if Gaussian densities are made correctly |
    # \------------------------------------------------/
    for gL in range(2, 9):
        g = gauss.get_gauss(gL)  # a gaussian of gL'th order
        print('\nGaussian of order', gL)
        for mL in range(9):
            m = gauss.get_moment(g, mL)  # the mL'th moment of g
            print('  %s\'th moment = %2.6f' % (mL, m))
            equal(m, gL == mL, 1e-4)

    # Check the moments of the constructed 1s density
    print('\nDensity of Hydrogen atom')
    for L in range(4):
        m = gauss.get_moment(nH, L)
        print('  %s\'th moment = %2.6f' % (L, m))
        equal(m, (L == 0) / sqrt(4 * pi), 1.5e-3)

    # Check that it is removed correctly
    gauss.remove_moment(nH, 0)
    m = gauss.get_moment(nH, 0)
    print('\nZero\'th moment of compensated Hydrogen density =', m)
    equal(m, 0., 1e-7)

    # /-------------------------------------------------\
    # | Check if Gaussian potentials are made correctly |
    # \-------------------------------------------------/

    # Array for storing the potential
    pot = gd.zeros(dtype=float, global_array=False)
    for L in range(7):  # Angular index of gaussian
        # Get analytic functions
        ng = gauss.get_gauss(L)
        vg = gauss.get_gauss_pot(L)

        # Solve potential numerically
        solve(pot, ng, charge=None, zero_initial_phi=True)

        # Determine residual
        residual = norm(pot - vg)
        residual = gd.integrate((pot - vg)**2)**0.5

        # print result
        print('L=%s, processor %s of %s: %s' % (
            L,
            gd.comm.rank + 1,
            gd.comm.size,
            residual))

        assert residual < 0.6

    # mpirun -np 2 python gauss_func.py --gpaw-parallel --gpaw-debug



from math import pi
from collections import defaultdict

import numpy as np

from gpaw.kpt_descriptor import KPointDescriptor
from .kpts import RSKPoint, to_real_space


def create_symmetry_map(kd: KPointDescriptor):  # -> List[List[int]]
    sym = kd.symmetry
    U_scc = sym.op_scc
    nsym = len(U_scc)
    compconj_s = np.zeros(nsym, bool)
    if sym.time_reversal and not sym.has_inversion:
        U_scc = np.concatenate([U_scc, -U_scc])
        compconj_s = np.zeros(nsym * 2, bool)
        compconj_s[nsym:] = True
        nsym *= 2

    map_ss = np.zeros((nsym, nsym), int)
    for s1 in range(nsym):
        for s2 in range(nsym):
            diff_s = abs(U_scc[s1].dot(U_scc).transpose((1, 0, 2)) -
                         U_scc[s2]).sum(2).sum(1)
            indices = (diff_s == 0).nonzero()[0]
            assert len(indices) == 1
            s = indices[0]
            assert compconj_s[s1] ^ compconj_s[s2] == compconj_s[s]
            map_ss[s1, s2] = s

    return map_ss


class Symmetry:
    def __init__(self, kd: KPointDescriptor):
        self.kd = kd
        self.symmetry_map_ss = create_symmetry_map(kd)

        U_scc = kd.symmetry.op_scc
        is_identity_s = (U_scc == np.eye(3, dtype=int)).all(2).all(1)
        self.s0 = is_identity_s.nonzero()[0][0]
        self.inverse_s = self.symmetry_map_ss[:, self.s0]

    def symmetry_operation(self, s: int, wfs, inverse=False):
        if inverse:
            s = self.inverse_s[s]
        U_scc = self.kd.symmetry.op_scc
        nsym = len(U_scc)
        time_reversal = s >= nsym
        s %= nsym
        U_cc = U_scc[s]

        if (U_cc == np.eye(3, dtype=int)).all():
            def T0(a_R):
                return a_R
        else:
            N_c = wfs.gd.N_c
            i_cr = np.dot(U_cc.T, np.indices(N_c).reshape((3, -1)))
            i = np.ravel_multi_index(i_cr, N_c, 'wrap')

            def T0(a_R):
                return a_R.ravel()[i].reshape(N_c)

        if time_reversal:
            def T(a_R):
                return T0(a_R).conj()
        else:
            T = T0

        T_a = []
        for a, id in enumerate(wfs.setups.id_a):
            b = self.kd.symmetry.a_sa[s, a]
            S_c = np.dot(wfs.spos_ac[a], U_cc) - wfs.spos_ac[b]
            U_ii = wfs.setups.atomrotations.get_by_a(a).R_sii[s].T
            T_a.append((b, S_c, U_ii))

        return T, T_a, time_reversal

    def apply_symmetry(self, s: int, rsk, wfs, spos_ac):
        U_scc = self.kd.symmetry.op_scc
        nsym = len(U_scc)
        time_reversal = s >= nsym
        s %= nsym
        sign = 1 - 2 * int(time_reversal)
        U_cc = U_scc[s]

        if (U_cc == np.eye(3)).all() and not time_reversal:
            return rsk

        u1_nR = rsk.u_nR
        proj1 = rsk.proj
        f_n = rsk.f_n
        k1_c = rsk.k_c
        weight = rsk.weight

        u2_nR = np.empty_like(u1_nR)
        proj2 = proj1.new()

        k2_c = sign * U_cc.dot(k1_c)

        N_c = u1_nR.shape[1:]
        i_cr = np.dot(U_cc.T, np.indices(N_c).reshape((3, -1)))
        i = np.ravel_multi_index(i_cr, N_c, 'wrap')
        for u1_R, u2_R in zip(u1_nR, u2_nR):
            u2_R[:] = u1_R.ravel()[i].reshape(N_c)

        for a, id in enumerate(wfs.setups.id_a):
            b = self.kd.symmetry.a_sa[s, a]
            S_c = np.dot(spos_ac[a], U_cc) - spos_ac[b]
            x = np.exp(2j * pi * np.dot(k1_c, S_c))
            U_ii = wfs.setups.atomrotations.get_by_a(a).R_sii[s].T * x
            proj2[a][:] = proj1[b].dot(U_ii)

        if time_reversal:
            np.conj(u2_nR, out=u2_nR)
            np.conj(proj2.array, out=proj2.array)

        return RSKPoint(u2_nR, proj2, f_n, k2_c, weight)

    def pairs(self, kpts, wfs, spos_ac):
        kd = self.kd
        nsym = len(kd.symmetry.op_scc)

        assert len(kpts) == kd.nibzkpts

        symmetries_k = []
        for k in range(kd.nibzkpts):
            indices = np.where(kd.bz2ibz_k == k)[0]
            sindices = (kd.sym_k[indices] +
                        kd.time_reversal_k[indices] * nsym)
            symmetries_k.append(sindices)

        # pairs: Dict[Tuple[int, int, int], int]

        pairs1 = defaultdict(int)
        for i1 in range(kd.nibzkpts):
            for s1 in symmetries_k[i1]:
                for i2 in range(kd.nibzkpts):
                    for s2 in symmetries_k[i2]:
                        s3 = self.symmetry_map_ss[s1, s2]
                        # s3 = self.inverse_s[s3]
                        if 1:  # i1 < i2:
                            pairs1[(i1, i2, s3)] += 1
                        else:
                            s4 = self.inverse_s[s3]
                            if i1 == i2:
                                # pairs1[(i1, i1, min(s3, s4))] += 1
                                pairs1[(i1, i1, s3)] += 1
                            else:
                                pairs1[(i2, i1, s4)] += 1
        pairs = {}
        seen = {}
        for (i1, i2, s), count in pairs1.items():
            k2 = kd.bz2bz_ks[kd.ibz2bz_k[i2], s]
            if (i1, k2) in seen:
                pairs[seen[(i1, k2)]] += count
            else:
                pairs[(i1, i2, s)] = count
                # seen[(i1, k2)] = (i1, i2, s)

        comm = wfs.world
        lasti1 = -1
        lasti2 = -1
        for (i1, i2, s), count in sorted(pairs.items()):
            if i1 != lasti1:
                k1 = kpts[i1]
                u1_nR = to_real_space(k1.psit)
                rsk1 = RSKPoint(u1_nR, k1.proj.broadcast(),
                                k1.f_n, k1.k_c,
                                k1.weight, k1.dPdR_aniv)
                lasti1 = i1
            if i2 == i1:
                if s == self.s0:
                    rsk2 = rsk1
                else:
                    N = len(rsk1.u_nR)
                    S = comm.size
                    B = (N + S - 1) // S
                    na = min(B * comm.rank, N)
                    nb = min(na + B, N)
                    rsk2 = RSKPoint(rsk1.u_nR[na:nb],
                                    rsk1.proj.view(na, nb),
                                    rsk1.f_n[na:nb],
                                    rsk1.k_c,
                                    rsk1.weight)
                lasti2 = i2
            elif i2 != lasti2:
                k2 = kpts[i2]
                N = len(k2.psit.array)
                S = comm.size
                B = (N + S - 1) // S
                na = min(B * comm.rank, N)
                nb = min(na + B, N)
                u2_nR = to_real_space(k2.psit, na, nb)
                rsk2 = RSKPoint(u2_nR, k2.proj.broadcast().view(na, nb),
                                k2.f_n[na:nb], k2.k_c,
                                k2.weight)
                lasti2 = i2

            yield (i1, i2, s, rsk1,
                   self.apply_symmetry(s, rsk2, wfs, spos_ac),
                   count)
